from py_wake.aep_calculator import AEPCalculator
from py_wake.examples.data.iea37._iea37 import IEA37_WindTurbines, IEA37Site
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.plotting import XYPlotComp
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
import topfarm
import numpy as np
from scipy.interpolate import RegularGridInterpolator
import warnings
from py_wake.flow_map import Points
from py_wake.utils.gradients import autograd


class PyWakeAEP(AEPCalculator):
    """TOPFARM wrapper for PyWake AEP calculator"""

    def __init__(self, wake_model):
        warnings.warn('PyWakeAEP is deprecated. Use PyWakeAEPCostModelComponent instead', DeprecationWarning)
        AEPCalculator.__init__(self, wake_model)

    def get_TopFarm_cost_component(self, n_wt, wd=None, ws=None):
        """Create topfarm-style cost component

        Parameters
        ----------
        n_wt : int
            Number of wind turbines
        """
        return AEPCostModelComponent(
            input_keys=['x', 'y'],
            n_wt=n_wt,
            cost_function=lambda **kwargs:
                self.calculate_AEP(x_i=kwargs[topfarm.x_key],
                                   y_i=kwargs[topfarm.y_key],
                                   h_i=kwargs.get(topfarm.z_key, None),
                                   type_i=kwargs.get(topfarm.type_key, 0),
                                   wd=wd, ws=ws).sum(),
            output_unit='GWh')

    def get_aep4smart_start(self, ws=[6, 8, 10, 12, 14], wd=np.arange(360)):
        def aep4smart_start(X, Y, wt_x, wt_y, T=0, wt_t=0):
            x = np.sort(np.unique(X))
            y = np.sort(np.unique(Y))
            X_j, Y_j, aep_map = self.aep_map(x, y, T, wt_x, wt_y, wd=wd, ws=ws, wt_type=wt_t)
#             import matplotlib.pyplot as plt
#             c = plt.contourf(X_j, Y_j, aep_map[:, :, 0], 100)
#             plt.colorbar(c)
#             plt.show()
            return RegularGridInterpolator((y, x), aep_map.values)(np.array([Y, X]).T)
        return aep4smart_start


class PyWakeAEPCostModelComponent(AEPCostModelComponent):
    """TOPFARM wrapper for PyWake AEP calculator"""

    def __init__(self, windFarmModel, n_wt, wd=None, ws=None, max_eval=None, grad_method=autograd, n_cpu=1, **kwargs):
        """Initialize wrapper for PyWake AEP calculator

        Parameters
        ----------
        windFarmModel : DeficitModel
            Wake deficit model used
        n_wt : int
            Number of wind turbines
        wd : array_like
            Wind directions to study
        ws : array_like
            Wind speeds to study
        max_eval : int
            Maximum number of function evaluations
        grad_method : function handle
            Selected method to calculate gradients, default is autograd
        """
        self.windFarmModel = windFarmModel
        self.n_cpu = n_cpu

        def aep(**kwargs):
            try:
                return self.windFarmModel.aep(x=kwargs[topfarm.x_key],
                                              y=kwargs[topfarm.y_key],
                                              h=kwargs.get(topfarm.z_key, None),
                                              type=kwargs.get(topfarm.type_key, 0),
                                              wd=wd, ws=ws,
                                              n_cpu=n_cpu)
            except ValueError as e:
                if 'are at the same position' in str(e):
                    return 0

        if grad_method:
            if hasattr(self.windFarmModel, 'dAEPdxy'):
                # for backward compatibility
                dAEPdxy = self.windFarmModel.dAEPdxy(grad_method)
            else:
                def dAEPdxy(**kwargs):
                    return self.windFarmModel.aep_gradients(
                        gradient_method=grad_method, wrt_arg=['x', 'y'], n_cpu=n_cpu, **kwargs)

            def daep(**kwargs):
                return dAEPdxy(x=kwargs[topfarm.x_key],
                               y=kwargs[topfarm.y_key],
                               h=kwargs.get(topfarm.z_key, None),
                               type=kwargs.get(topfarm.type_key, 0),
                               wd=wd, ws=ws)
        else:
            daep = None
        AEPCostModelComponent.__init__(self,
                                       input_keys=[topfarm.x_key, topfarm.y_key],
                                       n_wt=n_wt,
                                       cost_function=aep,
                                       cost_gradient_function=daep,
                                       output_unit='GWh',
                                       max_eval=max_eval, **kwargs)

    def get_aep4smart_start(self, ws=[6, 8, 10, 12, 14], wd=np.arange(360), type=0):
        """Compute AEP with a smart start approach"""

        def aep4smart_start(X, Y, wt_x, wt_y, T=0, wt_t=0):
            H = np.full(X.shape, self.windFarmModel.windTurbines.hub_height())
            if type == 0:
                sim_res = self.windFarmModel(wt_x, wt_y, type=wt_t, wd=wd, ws=ws, n_cpu=self.n_cpu)
                next_type = T
            else:
                type_ = np.atleast_1d(type)
                t = np.zeros_like(wt_x) + type_[:len(wt_x)]
                sim_res = self.windFarmModel(wt_x, wt_y, type=t, wd=wd, ws=ws, n_cpu=self.n_cpu)
                H = np.full(X.shape, self.windFarmModel.windTurbines.hub_height())
                next_type = type_[min(len(type_) - 1, len(wt_x) + 1)]
            return sim_res.aep_map(Points(X, Y, H), type=next_type, n_cpu=self.n_cpu).values
        return aep4smart_start


def main():
    if __name__ == '__main__':
        n_wt = 16
        site = IEA37Site(n_wt)
        windTurbines = IEA37_WindTurbines()
        windFarmModel = IEA37SimpleBastankhahGaussian(site, windTurbines)
        tf = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=PyWakeAEPCostModelComponent(windFarmModel, n_wt),
            driver=EasyRandomSearchDriver(randomize_func=RandomizeTurbinePosition_Circle(), max_iter=5),
            constraints=[CircleBoundaryConstraint([0, 0], 1300.1)],
            plot_comp=XYPlotComp())
        tf.optimize()
        tf.plot_comp.show()


main()



import numpy as np

from topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.plotting import XYPlotComp

from py_wake.deficit_models.gaussian import BastankhahGaussian
from py_wake.utils.gradients import autograd
from py_wake.validation.lillgrund import wt_x, wt_y, LillgrundSite, ct_curve, power_curve
from py_wake.wind_turbines import WindTurbine
from py_wake.wind_turbines.power_ct_functions import PowerCtTabular


def main():
    if __name__ == '__main__':
        wind_turbines = WindTurbine('SWT2p3_93_65', 92.6, 65, PowerCtTabular(ct_curve[:,0],power_curve[:,1], 'kW', ct_curve[:,1]))
        n_wt = len(wt_x)
        site = LillgrundSite()
        wf_model = BastankhahGaussian(site, wind_turbines)
        constraint_comp = XYBoundaryConstraint(np.asarray([wt_x, wt_y]).T)
        cost_comp = PyWakeAEPCostModelComponent(windFarmModel=wf_model,
                                                n_wt=n_wt,
                                                grad_method=autograd)

        problem = TopFarmProblem(design_vars={'x': wt_x, 'y': wt_y},
                                constraints=[constraint_comp, SpacingConstraint(min_spacing=wind_turbines.diameter() * 2)],
                                cost_comp=cost_comp,
                                driver=EasyScipyOptimizeDriver(optimizer='SLSQP', maxiter=5),
                                plot_comp=XYPlotComp())

        cost, state, recorder = problem.optimize(disp=True)

main()



import numpy as np
import matplotlib.pyplot as plt
from py_wake.examples.data.hornsrev1 import V80, Hornsrev1Site
from py_wake import BastankhahGaussian
from py_wake.utils.gradients import autograd
from topfarm import TopFarmProblem, TopFarmGroup
from topfarm.plotting import XYPlotComp
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.utils import regular_generic_layout, regular_generic_layout_gradients


def main():
    if __name__ == '__main__':
        site = Hornsrev1Site()
        wt = V80()
        D = wt.diameter()
        windFarmModel = BastankhahGaussian(site, wt)
        n_wt = 16
        stagger = 1 * D

        def reg_func(sx, sy, rotation, **kwargs):
            x, y = regular_generic_layout(n_wt, sx, sy, stagger, rotation)
            return [x, y]

        def reg_grad(sx, sy, rotation, **kwargs):
            dx_dsx, dy_dsx, dx_dsy, dy_dsy, dx_dr, dy_dr = regular_generic_layout_gradients(n_wt, sx, sy, stagger, rotation)
            return [[dx_dsx, dy_dsx], [dx_dsy, dy_dsy], [dx_dr, dy_dr]]

        reg_grid_comp = CostModelComponent(input_keys=[('sx', 0), 
                                                       ('sy', 0),
                                                       ('rotation', 0)],
                                      n_wt=n_wt,
                                      cost_function=reg_func,
                                      cost_gradient_function = reg_grad,
                                      output_keys= [('x', np.zeros(n_wt)), ('y', np.zeros(n_wt))],
                                      objective=False,
                                      use_penalty=False,
                                      )

        # AEP function 
        def aep_fun(x, y):
            aep = windFarmModel(x, y).aep().sum()
            return aep

        # predefinition of the AEP gradients 
        daep = windFarmModel.aep_gradients(gradient_method=autograd, wrt_arg=['x', 'y'])
        # AEP component
        aep_comp = CostModelComponent(input_keys=['x', 'y'],
                                      n_wt=n_wt,
                                      cost_function=aep_fun,
                                      cost_gradient_function = daep,
                                      output_keys= ("aep", 0),
                                      output_unit="GWh",
                                      maximize=True,
                                      objective=True)

        # OPTIMIZATION 
        plt.close('all')
        costs, states, recorders = [], [], []
        ec = 0.1
        boundary = [[-800,-50], [1200, -50], [1200,2300], [-800, 2300]]
        problem = TopFarmProblem(design_vars={'sx': (3*D, 2*D, 15*D), 
                                              'sy': (4*D, 2*D, 15*D), 
                                               'rotation': (50, 0, 90)
                                              },
                                 constraints=[XYBoundaryConstraint(boundary),
                                              SpacingConstraint(4*D)],
                                 grid_layout_comp=reg_grid_comp,
                                 n_wt = n_wt,
                                 cost_comp=aep_comp,
                                 driver=EasyScipyOptimizeDriver(optimizer='SLSQP', maxiter=200),
                                 plot_comp=XYPlotComp(),
                                 expected_cost=ec,
                                 )

        # problem.evaluate()
        cost, state, recorder = problem.optimize()

main()



import numpy as np
from topfarm.cost_models.cost_model_wrappers import CostModelComponent,\
    AEPCostModelComponent, AEPMaxLoadCostModelComponent
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint, CircleBoundaryComp, CircleBoundaryConstraint
from topfarm import TopFarmProblem
from topfarm.easy_drivers import EasyScipyOptimizeDriver, EasySimpleGADriver
from topfarm.plotting import NoPlot, XYPlotComp
from topfarm.tests import npt
from topfarm.tests.test_files import xy3tb


boundary = [(0, 0), (6, 0), (6, -10), (0, -10)]  # turbine boundaries
initial = np.array([[6, 0], [6, -8], [1, 1], [-1, -8]])  # initial turbine layouts
optimal_with_constraints = np.array([[2.5, -3], [6, -7], [4.5, -3], [3, -7]])  # optimal turbine layout
min_spacing = 2  # min distance between turbines
optimal = np.array([[3, -3], [7, -7], [4, -3], [3, -7]])  # desired turbine layouts


def get_tf(cost_comp, plot_comp=NoPlot()):
    return TopFarmProblem(
        dict(zip('xy', initial.T)),
        cost_comp=cost_comp,
        plot_comp=plot_comp,
        constraints=[SpacingConstraint(min_spacing), XYBoundaryConstraint(boundary)],
        driver=EasyScipyOptimizeDriver(disp=False),
        expected_cost=1e-1)


def cost(x, y):
    opt_x, opt_y = optimal.T
    return np.sum((x - opt_x)**2 + (y - opt_y)**2)


def aep_cost(x, y):
    opt_x, opt_y = optimal.T
    return -np.sum((x - opt_x)**2 + (y - opt_y)**2)


def gradients(x, y):
    return (2 * x - 2 * optimal[:, 0]), (2 * y - 2 * optimal[:, 1])


def aep_gradients(x, y):
    return -(2 * x - 2 * optimal[:, 0]), -(2 * y - 2 * optimal[:, 1])


def test_CostModelComponent():
    tf = get_tf(CostModelComponent(['x', 'y'], 4, cost, gradients))
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)


def testCostModelComponent_no_gradients():
    tf = get_tf(CostModelComponent(['x', 'y'], 4, cost, None))
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 4)


def testAEPCostModelComponent():
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients))
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)


def test_maxiter_CostModelComponent():
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients, max_eval=10))
    cost, state, recorder = tf.optimize()
    assert 10 <= tf.cost_comp.counter <= 12, tf.cost_comp.counter
    npt.assert_array_equal(recorder['AEP'][tf.cost_comp.n_func_eval], recorder['AEP'][tf.cost_comp.n_func_eval:])


def testCostModelComponentDiffShapeInput():
    def aep_cost(x, y, h):
        opt_x, opt_y = optimal.T
        return -np.sum((x - opt_x)**2 + (y - opt_y)**2) + h, {'add_out': sum(x)}
    cost_comp = AEPCostModelComponent(['x', 'y', ('h', 0)], 4, aep_cost, additional_output=[('add_out', 0)])
    tf = TopFarmProblem(
        dict(zip('xy', initial.T)),
        cost_comp=cost_comp,
        constraints=[SpacingConstraint(min_spacing), XYBoundaryConstraint(boundary)],
        driver=EasyScipyOptimizeDriver(disp=False),
        ext_vars={'h': 0})
    cost0, _, _ = tf.optimize(state={'h': 0})
    cost10, _, _ = tf.optimize(state={'h': 10})
    npt.assert_almost_equal(cost10, cost0 - 10)


def testCostModelComponentAdditionalOutput():
    def aep_cost(x, y):
        opt_x, opt_y = optimal.T
        return -np.sum((x - opt_x)**2 + (y - opt_y)**2), {'add_out': sum(x)}
    tf = get_tf(AEPCostModelComponent(['x', 'y'], 4, aep_cost, aep_gradients, additional_output=[('add_out', 0)]))
    _, state, _ = tf.optimize()
    npt.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal_with_constraints, 5)
    npt.assert_equal(sum(state['x']), state['add_out'])


def test_AEPMaxLoadCostModelComponent_as_penalty():
    tf = xy3tb.get_tf(
        design_vars={'x': ([0]), 'y': ([0])},
        cost_comp=AEPMaxLoadCostModelComponent(
            input_keys='xy', n_wt=1,
            aep_load_function=lambda x, y: (-np.sin(np.hypot(x, y)), np.hypot(x, y)),
            max_loads=3),
        constraints=[],
        driver=EasySimpleGADriver(),
        plot_comp=None)

    # check normal result that satisfies the penalty
    assert tf.evaluate({'x': np.pi / 2})[0] == 1
    # check penalized result if capacity constraint is not satisfied
    for x, y in [(4, 0), (0, 4)]:
        assert tf.evaluate({'x': x, 'y': y})[0] == 1e10 + 1


def test_AEPMaxLoadCostModelComponent_as_penalty_multi_wt():
    tf = xy3tb.get_tf(
        design_vars={'x': ([0, 1]), 'y': ([0, 0])},
        cost_comp=AEPMaxLoadCostModelComponent(
            input_keys='xy', n_wt=2,
            output_keys=["AEP", ('loads', [3, 3])],
            aep_load_function=lambda x, y: (-np.sin(np.hypot(x, y)).sum(), np.hypot(x, y)),
            max_loads=[3, 3]),
        constraints=[],
        driver=EasySimpleGADriver(),
        plot_comp=None)

    # check normal result that satisfies the penalty
    assert tf.evaluate({'x': [np.pi / 2, -np.pi / 2]})[0] == 2
    # check penalized result if capacity constraint is not satisfied
    for x, y in [([4, 0], [0, 0]), ([0, 0], [4, 0]), ([4, 4], [0, 0])]:
        assert tf.evaluate({'x': x, 'y': y})[0] == 1e10 + 1


def test_AEPMaxLoadCostModelComponent_constraint():

    tf = TopFarmProblem(
        design_vars={'x': ([1]), 'y': (.1, 0, 2.5)},
        # design_vars={'x': ([2.9], [1], [3])},
        cost_comp=AEPMaxLoadCostModelComponent(
            input_keys='xy', n_wt=1,
            aep_load_function=lambda x, y: (np.hypot(x, y), x),
            max_loads=3),
        constraints=[CircleBoundaryConstraint((0, 0), 7)],
    )

    tf.evaluate()
    cost, state, recorder = tf.optimize()
    npt.assert_allclose(state['x'], 3)  # constrained by max_loads
    npt.assert_allclose(state['y'], 2.5)  # constrained by design var lim



import numpy as np
from topfarm._topfarm import TopFarmProblem, TopFarmGroup, TopFarmParallelGroup
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.plotting import NoPlot
import time
from openmdao.api import n2



from openmdao.api import Problem, IndepVarComp, ParallelGroup, ExecComp, Group

def main():
    if __name__ == '__main__':
        # define the conditions for the wind farm
        boundary = [(0, 0), (6, 0), (6, -10), (0, -10)]  # turbine boundaries
        initial = np.array([[6, 0], [6, -8], [1, 1], [-1, -8]])  # initial turbine pos
        desired = np.array([[3, -3], [7, -7], [4, -3], [3, -7]])  # desired turbine pos
        optimal = np.array([[2.5, -3], [6, -7], [4.5, -3], [3, -7]])  # optimal layout
        min_spacing = 2  # min distance between turbines

        # ------------------------ OPTIMIZATION ------------------------

        # create the wind farm and run the optimization

        def wt_cost(i, x, y):
            time.sleep(0.01)
            return (desired[i, 0] - x[i])**2 + (desired[i, 1] - y[i])**2

        n_wt = len(initial)
        comps = [CostModelComponent('xy', 4,
                                    cost_function=lambda x, y, i=i:wt_cost(i, x, y),
                                    objective=False,
                                    output_keys='cost%d' % i) for i in range(n_wt)]

        def sum_map(**kwargs):

            return np.sum([kwargs['cost%d' % i] for i in range(n_wt)])

        comps.append(CostModelComponent(['cost%d' % i for i in range(n_wt)], 1,
                                        cost_function=sum_map,
                                        objective=True))
        cost_comp = TopFarmParallelGroup(comps)

        tf = TopFarmProblem(
            design_vars={'x': initial[:, 0], 'y': initial[:, 1]},
            cost_comp=cost_comp,
            constraints=[XYBoundaryConstraint(boundary),
                         SpacingConstraint(min_spacing)],
#            plot_comp=DummyCostPlotComp(desired),
            plot_comp=NoPlot(),
            driver=EasyScipyOptimizeDriver()
        )
#        n2(tf)
        #print(tf.evaluate({'x': desired[:, 0], 'y': desired[:, 1]}))
        print(tf.evaluate({'x': optimal[:, 0], 'y': optimal[:, 1]}, disp=False))
        #print(tf.evaluate({'x': initial[:, 0], 'y': initial[:, 1]}))
#        tic = time.time()
        cost, state, recorder = tf.optimize()
#        toc = time.time()
#        print('optimized in {:.3f}s '.format(toc-tic))
        tf.plot_comp.show()


main()
N_PROCS = 2



# -*- coding: utf-8 -*-
"""
Created on Thu Oct  6 13:55:21 2022

@author: mikf
"""
import numpy as np
from topfarm.constraint_components import Constraint, ConstraintComponent
from topfarm.cost_models.cost_model_wrappers import CostModelComponent


class ConstraintAggregation(Constraint):
    def __init__(self, constraints, **kwargs):
        self.constraints = constraints
        self.const_id = 'constraint_aggregation_comp'
        self.kwargs = kwargs
        self.input_keys = list([(i, i[0])[isinstance(i, tuple)] for i in kwargs['component_args']['input_keys']])
        self.output_keys = list([(i, i[0])[isinstance(i, tuple)] for i in kwargs['component_args']['output_keys']])

    @property
    def constraintComponent(self):
        return self.constraint_aggregation_comp

    def _setup(self, problem):
        self.constraint_aggregation_comp = ConstraintAggregationComp(**self.kwargs['component_args'])
        for constraint in self.constraints:
            constraint._setup(problem)
        problem.model.add_subsystem(self.const_id, self.constraint_aggregation_comp,
                                    promotes=['*'])

    def setup_as_constraint(self, problem):
        self._setup(problem)
        problem.model.add_constraint(**self.kwargs['constraint_args'])

    def setup_as_penalty(self, problem):
        self._setup(problem)


class ConstraintAggregationComp(CostModelComponent, ConstraintComponent):
    def __init__(self, **kwargs):
        CostModelComponent.__init__(self, **kwargs)

    def satisfy(self, state):
        pass


class DistanceConstraintAggregation(ConstraintAggregation):
    """Aggregating the spacing and boundary distances constraints into a penalty function"""

    def __init__(self, constraints, n_wt, min_spacing_m, windTurbines, name='sgd_constraint', **kwargs):
        """Initializing spacing constraint aggregation class

        Parameters
        ----------
        constraints : list
            list of constraint components, common constraints are [SpacingConstraints] and [XYBoundaryConstraints]
        n_wt : int
            number of turbines
        min_spacing_m : float
            minimum inter turbine spacing in meters
        windTurbines : object
            pywake object for the wind turbine used
        """

        def constr_aggr_func(wtSeparationSquared, boundaryDistances, **kwargs):
            separation_constraint = wtSeparationSquared - (2 * windTurbines.diameter()) ** 2
            separation_constraint = separation_constraint[separation_constraint < 0]
            distance_constraint = boundaryDistances
            distance_constraint = distance_constraint[distance_constraint < 0]
            return np.sum(-1 * separation_constraint) + np.sum(distance_constraint ** 2)

        def constr_aggr_grad(wtSeparationSquared, boundaryDistances, **kwargs):
            separation_constraint = wtSeparationSquared - (2 * windTurbines.diameter()) ** 2
            J_separation = np.zeros_like(wtSeparationSquared)
            J_separation[np.where(separation_constraint < 0)] = -1
            J_distance = np.zeros_like(boundaryDistances)
            J_distance[np.where(boundaryDistances < 0)] = 2 * boundaryDistances[np.where(boundaryDistances < 0)]
            return [[J_separation], [J_distance]]

        kwargs['component_args'] = {'input_keys': [('wtSeparationSquared', np.zeros(int(n_wt * (n_wt - 1) / 2))),
                                                   ('boundaryDistances', np.zeros((n_wt, 4)))],
                                    'n_wt': n_wt,
                                    'cost_function': constr_aggr_func,
                                    'cost_gradient_function': constr_aggr_grad,
                                    'objective': False,
                                    'output_keys': [(name, 0)],
                                    'use_penalty': False}
        kwargs['constraint_args'] = {'name': name, 'lower': 0}

        ConstraintAggregation.__init__(self, constraints, **kwargs)



"""Example: optimization with different wake models

This example uses a dummy cost function to optimize a simple wind turbine
layout that is subject to constraints. The optimization pushes the wind turbine
locations to specified locations in the farm.
"""
import os

from matplotlib.patches import Polygon
import numpy as np
from py_wake.deficit_models.gcl import GCL
from py_wake.deficit_models.noj import NOJ
from py_wake.examples.data.hornsrev1 import V80
from py_wake.site._site import UniformWeibullSite
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.plotting import XYPlotComp, NoPlot


def main():
    if __name__ == '__main__':
        try:
            import matplotlib.pyplot as plt
            plt.gcf()
            plot_comp = XYPlotComp
            plot = True
        except RuntimeError:
            plot_comp = NoPlot
            plot = False
        # ------------------------ INPUTS ------------------------

        # ------------------------ DEFINE WIND RESOURCE ------------------------
        # wind resource info (wdir frequency, weibull a and k)
        f = [3.597152, 3.948682, 5.167395, 7.000154, 8.364547, 6.43485, 8.643194,
             11.77051, 15.15757, 14.73792, 10.01205, 5.165975]
        a = [9.176929, 9.782334, 9.531809, 9.909545, 10.04269, 9.593921, 9.584007,
             10.51499, 11.39895, 11.68746, 11.63732, 10.08803]
        k = [2.392578, 2.447266, 2.412109, 2.591797, 2.755859, 2.595703, 2.583984,
             2.548828, 2.470703, 2.607422, 2.626953, 2.326172]
        site = UniformWeibullSite(p_wd=f, a=a, k=k, ti=0.075)
        wt = V80()

        # ------------------------ setup problem ____---------------------------
        rot_diam = 80.0  # rotor diameter [m]
        init_pos = np.array([(0, 2 * rot_diam), (0, 0),
                             (0, -2 * rot_diam)])  # initial turbine positions
        b = 2 * rot_diam + 10  # boundary size
        boundary = [(-b, -b), (-b, b), (b, b), (b, -b)]  # corners of wind farm boundary
        min_spacing = 2.0 * rot_diam  # minimum spacing between turbines [m]

        # ------------------------ OPTIMIZATION ------------------------

        def get_tf(windFarmModel):
            return TopFarmProblem(
                design_vars=dict(zip('xy', init_pos.T)),
                cost_comp=PyWakeAEPCostModelComponent(windFarmModel, n_wt=3, ws=10, wd=np.arange(0, 360, 12),
                                                      grad_method=None),
                constraints=[SpacingConstraint(min_spacing),
                             XYBoundaryConstraint(boundary)],
                driver=EasyScipyOptimizeDriver(),
                plot_comp=plot_comp(),
                expected_cost = 1e-4,
                )

        # GCL: define the wake model and optimization problem
        tf_gcl = get_tf(GCL(site, wt))

        # NOJ: define the wake model and optimization problem
        tf_noj = get_tf(NOJ(site, wt))

        # run the optimization
        cost_gcl, state_gcl, recorder_gcl = tf_gcl.optimize()
        cost_noj, state_noj, recorder_noj = tf_noj.optimize()

        # ------------------------ POST-PROCESS ------------------------

        # get the optimized locations
        opt_gcl = tf_gcl.turbine_positions
        opt_noj = tf_noj.turbine_positions

        # create the array of costs for easier printing
        costs = np.diag([cost_gcl, cost_noj])
        costs[0, 1] = tf_noj.evaluate(state_gcl)[0]  # noj cost of gcl locs
        costs[1, 0] = tf_gcl.evaluate(state_noj)[0]  # gcl cost of noj locs

        # ------------------------ PRINT STATS ------------------------

        aep_diffs = 200 * (costs[:, 0] - costs[:, 1]) / (costs[:, 0] + costs[:, 1])
        loc_diffs = 200 * (costs[0, :] - costs[1, :]) / (costs[0, :] + costs[1, :])

        print('\nComparison of cost models vs. optimized locations:')
        print('\nCost    |    GCL_aep      NOJ_aep')
        print('---------------------------------')
        print(f'GCL_loc |{costs[0,0]:11.2f} {costs[0,1]:11.2f}' +
              f'   ({aep_diffs[0]:.2f}%)')
        print(f'NOJ_loc |{costs[1,0]:11.2f} {costs[1,1]:11.2f}' +
              f'   ({aep_diffs[1]:.2f}%)')
        print(f'             ({loc_diffs[0]:.2f}%)     ({loc_diffs[1]:.2f}%)')

        # ------------------------ PLOT (if possible) ------------------------
        if plot:

            # initialize the figure and axes
            fig = plt.figure(1, figsize=(7, 5))
            plt.clf()
            ax = plt.axes()

            # plot the boundary and desired locations
            ax.add_patch(Polygon(boundary, fill=False,
                                 label='Boundary'))  # boundary
            ax.plot(init_pos[:, 0], init_pos[:, 1], 'xk',
                    label='Initial')
            ax.plot(opt_gcl[:, 0], opt_gcl[:, 1], 'o',
                    label='GCL')
            ax.plot(opt_noj[:, 0], opt_noj[:, 1], '^',
                    label='NOJ')

            # make a few adjustments to the plot
            ax.autoscale_view()  # autoscale the boundary
            plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                       ncol=4, mode='expand', borderaxespad=0.)  # add a legend
            plt.tight_layout()  # zoom the plot in
            plt.axis('off')  # remove the axis

            # save the png
            folder, file = os.path.split(__file__)
            fig.savefig(folder + "/figures/" + file.replace('.py', '.png'))

main()



"""Example: optimizing a layout with constraints

This example uses a dummy cost function to optimize a simple wind turbine
layout that is subject to constraints. The optimization pushes the wind turbine
locations to specified locations in the farm.
"""
import os

from matplotlib.patches import Polygon
import matplotlib.pyplot as plt
import numpy as np

from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.plotting import NoPlot


def main():
    if __name__ == '__main__':
        # ------------------------ INPUTS ------------------------

        # define the conditions for the wind farm
        boundary = [(0, 0), (6, 0), (6, -10), (0, -10)]  # turbine boundaries
        initial = np.array([[6, 0], [6, -8], [1, 1], [-1, -8]])  # initial turbine pos
        desired = np.array([[3, -3], [7, -7], [4, -3], [3, -7]])  # desired turbine pos
        optimal = np.array([[2.5, -3], [6, -7], [4.5, -3], [3, -7]])  # optimal layout
        min_spacing = 2  # min distance between turbines

        try:
            import matplotlib.pyplot as plt
            plt.gcf()
            plot_comp = DummyCostPlotComp(desired)
            plot = True
        except RuntimeError:
            plot_comp = NoPlot()
            plot = False
        # ------------------------ OPTIMIZATION ------------------------

        # create the wind farm and run the optimization

        tf = TopFarmProblem(
            design_vars={'x': initial[:, 0], 'y': initial[:, 1]},
            cost_comp=DummyCost(desired, ['x', 'y']),
            constraints=[XYBoundaryConstraint(boundary),
                         SpacingConstraint(min_spacing)],
            plot_comp=plot_comp,
            driver=EasyScipyOptimizeDriver()
        )
        cost, state, recorder = tf.optimize()
        tf.plot_comp.show()

        # final position
        final_x, final_y = state['x'], state['y']

        # get the positions tried during optimization from the recorder
        rec_x, rec_y = recorder['x'], recorder['y']

        # get the final, optimal positions
        optimized = tf.turbine_positions

        # ------------------------ PLOT (if possible) ------------------------

        if plot:

            # initialize the figure and axes
            fig = plt.figure(1, figsize=(7, 5))
            plt.clf()
            ax = plt.axes()

            # plot the boundary and desired locations
            ax.add_patch(Polygon(boundary, closed=True, fill=False,
                                 label='Boundary'))  # boundary
            plt.plot(desired[:, 0], desired[:, 1], 'ok', mfc='None', ms=10,
                     label='Desired')  # desired positions

            # plot the history of each turbine
            for i_turb in range(rec_x.shape[1]):
                l, = plt.plot(rec_x[0, i_turb], rec_y[0, i_turb],
                              'x', ms=8, label=f'Turbine {i_turb+1}')  # initial
                plt.plot(rec_x[:, i_turb], rec_y[:, i_turb],
                         c=l.get_color())  # tested values
                plt.plot(rec_x[-1, i_turb], rec_y[-1, i_turb],
                         'o', ms=8, c=l.get_color())  # final

            # make a few adjustments to the plot
            ax.autoscale_view()  # autoscale the boundary
            plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
                       ncol=4, mode='expand', borderaxespad=0.)  # add a legend
            plt.tight_layout()  # zoom the plot in
            plt.axis('off')  # remove the axis

            # save the png
            folder, file = os.path.split(__file__)
            fig.savefig(folder + "/figures/" + file.replace('.py', '.png'))

#        except RuntimeError:
#            pass


main()



import numpy as np
from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp

from topfarm.plotting import NoPlot, XYPlotComp
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.boundary import XYBoundaryConstraint,\
    PolygonBoundaryComp, InclusionZone, ExclusionZone, MultiPolygonBoundaryComp
from topfarm._topfarm import TopFarmProblem


def get_tf(initial, optimal, boundary, plot_comp=NoPlot(), boundary_type='polygon'):
    initial, optimal = map(np.array, [initial, optimal])
    return TopFarmProblem(
        {'x': initial[:, 0], 'y': initial[:, 1]},
        DummyCost(optimal),
        constraints=[XYBoundaryConstraint(boundary, boundary_type)],
        driver=EasyScipyOptimizeDriver(tol=1e-8, disp=False),
        plot_comp=plot_comp)


def testPolygon():
    boundary = [(0, 0), (1, 1), (2, 0), (2, 2), (0, 2)]
    b = PolygonBoundaryComp(0, boundary)
    np.testing.assert_array_equal(b.xy_boundary[:, :2], [[0, 0],
                                                         [1, 1],
                                                         [2, 0],
                                                         [2, 2],
                                                         [0, 2],
                                                         [0, 0]])


def testPolygonConcave():
    optimal = [(1.5, 1.3), (4, 1)]
    boundary = [(0, 0), (5, 0), (5, 2), (3, 2), (3, 1), (2, 1), (2, 2), (0, 2), (0, 0)]
    plot_comp = NoPlot()  # DummyCostPlotComp(optimal)
    initial = [(-0, .1), (4, 1.5)][::-1]
    tf = get_tf(initial, optimal, boundary, plot_comp)
    tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)
    plot_comp.show()


def testPolygonTwoRegionsStartInWrong():
    optimal = [(1, 1), (4, 1)]
    boundary = [(0, 0), (5, 0), (5, 2), (3, 2), (3, 0), (2, 0), (2, 2), (0, 2), (0, 0)]
    plot_comp = NoPlot()  # DummyCostPlotComp(optimal, delay=.1)
    initial = [(3.5, 1.5), (0.5, 1.5)]
    tf = get_tf(initial, optimal, boundary, plot_comp)
    tf.optimize()
    plot_comp.show()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)


def testMultiPolygon():
    optimal = [(1.75, 1.3), (4, 1)]
    boundary = [InclusionZone([(0, 0), (5, 0), (5, 2), (3, 2), (3, 1), (2, 1), (2, 2), (0, 2), (0, 0)]),
                InclusionZone([(3.5, 0.5), (4.5, 0.5), (4.5, 1.5), (3.5, 1.5)]),
                ExclusionZone([(0.5, 0.5), (1.75, 0.5), (1.75, 1.5), (0.5, 1.5)]),
                ExclusionZone([(0.75, 0.75), (1.25, 0.75), (1.25, 1.25), (0.75, 1.25)]),
                ]
    xy_bound_ref_ = np.array([[0., 0.],
                              [5., 0.],
                              [5., 2.],
                              [3., 2.],
                              [3., 1.],
                              [2., 1.],
                              [2., 2.],
                              [0., 2.],
                              [0., 0.]])

    bound_dist_ref = np.array([0, 1])
    plot_comp = NoPlot()  # DummyCostPlotComp(optimal)
    initial = np.asarray([(-0, .1), (4, 1.5)][::-1])
    tf = get_tf(initial, optimal, boundary, plot_comp, boundary_type='multi_polygon')
    tf.evaluate()
    cost, state, recorder = tf.optimize()
    np.testing.assert_array_almost_equal(recorder['xy_boundary'][-1], xy_bound_ref_, 4)
    np.testing.assert_array_almost_equal(recorder['boundaryDistances'][-1], bound_dist_ref, 4)
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)


def test_calculate_distance_to_boundary():
    import matplotlib.pyplot as plt
    boundary = np.array([(0, 0), (10, 0), (20, 10), (20, 20), (0, 20)])
    points = np.array([(2, 10), (10, 21), (14, 6)])

    boundary_constr = XYBoundaryConstraint(boundary, 'convex_hull').get_comp(10)
    import numpy.testing as npt
    if 0:
        plt.plot(*boundary.T, )
        plt.plot(*points.T, '.')
        plt.axis('equal')
        plt.grid()
        plt.show()
    npt.assert_array_almost_equal(boundary_constr.calculate_distance_to_boundary(points),
                                  [[18., 10., 2., 10., 12.72792206],
                                   [10., -1., 10., 21., 14.8492424],
                                   [6., 14., 14., 6., 1.41421356]])


def testDistanceRelaxation():
    boundary = [InclusionZone([(0, 0), (5, 0), (5, 2), (3, 2), (3, 1), (2, 1), (2, 2), (0, 2), (0, 0)]),
                InclusionZone([(3.5, 0.5), (4.5, 0.5), (4.5, 1.5), (3.5, 1.5)]),
                ExclusionZone([(0.5, 0.5), (1.75, 0.5), (1.75, 1.5), (0.5, 1.5)]),
                ExclusionZone([(0.75, 0.75), (1.25, 0.75), (1.25, 1.25), (0.75, 1.25)]),
                ]
    initial = [(-0, .1), (4, 1.5)][::-1]
    optimal = [(1.75, 1.3), (4, 1)]
    initial, optimal = map(np.array, [initial, optimal])
    plot_comp = NoPlot()
    tf = TopFarmProblem({'x': initial[:, 0], 'y': initial[:, 1]}, DummyCost(optimal, inputs=['x', 'y']),
                        constraints=[XYBoundaryConstraint(boundary, 'multi_polygon', relaxation=(0.9, 4))],
                        plot_comp=plot_comp, driver=EasyScipyOptimizeDriver(tol=1e-8, disp=False))
    tf.evaluate()
    cost, state, recorder = tf.optimize()
    np.testing.assert_array_almost_equal(tf.turbine_positions[:, :2], optimal, 4)
    relaxation = tf.model.constraint_components[0].calc_relaxation() \
        + tf.model.constraint_components[0].relaxation[0]
    assert tf.cost_comp.n_grad_eval <= 10
    assert tf.model.pre_constraints.xy_bound_comp == tf.model.constraint_components[0]
    # distances in the 2 lower corners should be the same
    assert tf.model.constraint_components[0].distances(np.array([0]), np.array([0])) \
        == tf.model.constraint_components[0].distances(np.array([5]), np.array([0]))
    # gradients with respect of iteration number should be the same at every point
    assert tf.model.constraint_components[0].gradients(np.array([3]), np.array([5]))[2][0] \
        == tf.model.constraint_components[0].gradients(np.array([1.5]), np.array([8]))[2][1]


def testDistanceRelaxationPolygons():
    zones = [InclusionZone([(0, 0), (5, 0), (5, 2), (3, 2), (3, 1), (2, 1), (2, 2), (0, 2), (0, 0)]),
             InclusionZone([(3.5, 0.5), (4.5, 0.5), (4.5, 1.5), (3.5, 1.5)]),
             ExclusionZone([(0.5, 0.5), (1.75, 0.5), (1.75, 1.5), (0.5, 1.5)]),
             ExclusionZone([(0.75, 0.75), (1.25, 0.75), (1.25, 1.25), (0.75, 1.25)]),
             ]
    MPBC = MultiPolygonBoundaryComp(1, zones, relaxation=(0.1, 10))
    (rp1, rp2) = MPBC.relaxed_polygons(7)
    bp1 = MPBC.get_boundary_properties(rp1)
    bp2 = MPBC.get_boundary_properties(rp2)
    np.testing.assert_allclose(bp1[0][0], np.array([[2.3, 1.3],
                                                    [2.7, 1.3],
                                                    [2.7, 2.3],
                                                    [5.3, 2.3],
                                                    [5.3, -0.3],
                                                    [-0.3, -0.3],
                                                    [-0.3, 2.3],
                                                    [2.3, 2.3]]))
    np.testing.assert_allclose(bp2[0][0], np.array([[1.45, 1.2],
                                                    [0.8, 1.2],
                                                    [0.8, 0.8],
                                                    [1.45, 0.8]]))



from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm.plotting import NoPlot
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.tests.test_files import xy3tb
from topfarm.constraint_components.spacing import SpacingConstraint, SpacingComp
from topfarm.tests import npt
from openmdao.utils.assert_utils import assert_check_partials
from openmdao.api import Problem, IndepVarComp
import numpy as np
from topfarm.utils import SmoothMin, LogSumExpMin, StrictMin
import pytest
from topfarm._topfarm import TopFarmProblem


@pytest.mark.parametrize('aggfunc', [  # None,
    StrictMin(),
    SmoothMin(.1),
    SmoothMin(1),
    SmoothMin(10),
    LogSumExpMin(.1),
    LogSumExpMin(1),
    LogSumExpMin(10)
])
@pytest.mark.parametrize('x,y', [([6, 5, -8, 1], [0, -8, -4, 1]),
                                 ([2.84532167, 7.00331189, 3.86523273], [-2.98101466, -6.99667302, -2.98433316])])
@pytest.mark.parametrize('full_aggregation', [True, False])
def test_spacing_4wt_partials(aggfunc, full_aggregation, x, y):

    from topfarm.constraint_components.boundary import XYBoundaryConstraint
    from topfarm.easy_drivers import EasyScipyOptimizeDriver
    import topfarm
    initial = desired = np.array([x, y]).T
    boundary = np.array([(0, 0), (6, 0), (6, -10), (0, -10)])  # turbine boundaries
    spacing_constr = SpacingConstraint(2, aggregation_function=aggfunc, full_aggregation=full_aggregation)

    k = {'cost_comp': DummyCost(desired[:, :2], [topfarm.x_key, topfarm.y_key]),
         'design_vars': {topfarm.x_key: initial[:, 0], topfarm.y_key: initial[:, 1]},
         'driver': EasyScipyOptimizeDriver(disp=True, tol=1e-8),
         'plot_comp': NoPlot(),
         'constraints': [spacing_constr, XYBoundaryConstraint(boundary)]}
    if 0:
        k['plot_comp'] = DummyCostPlotComp(desired)
    TopFarmProblem(**k)
    scomp = spacing_constr.constraintComponent
    outputs = {}

    def compute(x, y):
        scomp.compute(dict(x=x, y=y), outputs)
        return outputs['wtSeparationSquared']

    ref = compute(initial[:, 0], initial[:, 1])
    ddx = np.array([(compute(x, initial[:, 1]) - ref) / 1e-6 for x in initial[:, 0] + np.eye(len(x)) * 1e-6]).T
    ddy = np.array([(compute(initial[:, 0], y) - ref) / 1e-6 for y in initial[:, 1] + np.eye(len(x)) * 1e-6]).T

    scomp.compute_partials(dict(x=initial[:, 0], y=initial[:, 1]), outputs)
    npt.assert_array_almost_equal(outputs[('wtSeparationSquared', 'x')].reshape(ddx.shape), ddx, 4)
    npt.assert_array_almost_equal(outputs[('wtSeparationSquared', 'y')].reshape(ddy.shape), ddy, 4)


@pytest.mark.parametrize('aggfunc', [None,
                                     StrictMin(),
                                     SmoothMin(.1),
                                     SmoothMin(.5),
                                     LogSumExpMin(.1),
                                     LogSumExpMin(1),
                                     LogSumExpMin(10)
                                     ])
def test_spacing_4wt(aggfunc):

    from topfarm.constraint_components.boundary import XYBoundaryConstraint
    from topfarm.easy_drivers import EasyScipyOptimizeDriver
    import topfarm
    initial = np.array([[6, 0], [5, -8], [-1, -4], [1, 1]])  # initial turbine layouts
    desired = np.array([[3, -3], [7, -7], [3, -4], [4, -3]])  # desired turbine layouts
    boundary = np.array([(0, 0), (6, 0), (6, -10), (0, -10)])  # turbine boundaries
    # initial = np.array([[6, 0], [1, 1]])  # initial turbine layouts
    # desired = np.array([[3, -3], [4, -3]])  # desired turbine layouts

    k = {'cost_comp': DummyCost(desired[:, :2], [topfarm.x_key, topfarm.y_key]),
         'design_vars': {topfarm.x_key: initial[:, 0], topfarm.y_key: initial[:, 1]},
         'driver': EasyScipyOptimizeDriver(disp=True, tol=1e-8),
         'plot_comp': NoPlot(),
         'constraints': [SpacingConstraint(2, aggregation_function=aggfunc), XYBoundaryConstraint(boundary)]}
    if 0:
        k['plot_comp'] = DummyCostPlotComp(desired)
    tf = TopFarmProblem(**k)
    print(str(aggfunc))
    tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tf.plot_comp.show()
    tol = 1e-4
    assert sum((tb_pos[2] - tb_pos[0])**2) > 2**2 - tol  # check min spacing


@pytest.mark.parametrize('aggfunc', [
    # None,
    StrictMin(), SmoothMin(1), LogSumExpMin(1)])
def test_spacing(aggfunc):
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2, aggregation_function=aggfunc)], plot=False)
    tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tf.plot_comp.show()
    tol = 1e-4
    assert sum((tb_pos[2] - tb_pos[0])**2) > 2**2 - tol  # check min spacing


def test_spacing_as_penalty():
    driver = SimpleGADriver()
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2)],
                      driver=driver)

    # check normal result if spacing constraint is satisfied
    assert tf.evaluate()[0] == 45
    # check penalized result if spacing constraint is not satisfied
    assert tf.evaluate({'x': [3, 7, 4.], 'y': [-3., -7., -3.], 'z': [0., 0., 0.]})[0] == 1e10 + 3


def test_satisfy():
    sc = SpacingComp(n_wt=3, min_spacing=2)
    state = sc.satisfy(dict(zip('xy', xy3tb.desired.T)))
    x, y = state['x'], state['y']
    npt.assert_array_less(y, x)


def test_satisfy2():
    n_wt = 5
    sc = SpacingComp(n_wt=n_wt, min_spacing=2)
    theta = np.linspace(0, 2 * np.pi, n_wt, endpoint=False)
    x0, y0 = np.cos(theta), np.sin(theta)

    state = sc.satisfy({'x': x0, 'y': y0})
    x1, y1 = state['x'], state['y']
    if 0:
        import matplotlib.pyplot as plt
        colors = ['b', 'r', 'm', 'c', 'g', 'y', 'orange', 'indigo', 'grey']
        for i, (x0_, y0_, x1_, y1_) in enumerate(zip(x0, y0, x1, y1)):
            c = colors[i]
            plt.plot([x0_], [y0_], '>', color=c)
            plt.plot([x0_, x1_], [y0_, y1_], '-', color=c, label=i)
            plt.plot([x1_], [y1_], '.', color=c)
        plt.axis('equal')
        plt.legend()
        plt.show()

    dist = np.sqrt(sc._compute(x1, y1))
    npt.assert_array_less(2, dist)


@pytest.mark.parametrize('aggfunc', [None, StrictMin(), SmoothMin(1), LogSumExpMin(1)])
def test_partials(aggfunc):
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(2, aggregation_function=aggfunc)])
    # if complex numbers work: uncomment tf.setup below and
    # change method='cs' and step=1e-40 in check_partials
    # tf.setup(force_alloc_complex=True)

    # run to get rid of zeros initializaiton, otherwise not accurate partials
    tf.run_model()
    check = tf.check_partials(compact_print=True,
                              includes='spacing*',
                              method='fd',
                              step=1e-6,
                              form='central')
    atol = 1.e-6
    rtol = 1.e-6
    try:
        assert_check_partials(check, atol, rtol)
    except ValueError as err:
        print(str(err))
        raise


@pytest.mark.parametrize('aggfunc', [None,
                                     # StrictMin(), # not working
                                     SmoothMin(1), LogSumExpMin(1)])
def test_partials_many_turbines(aggfunc):
    n_wt = 10
    theta = np.linspace(0, 2 * np.pi, n_wt, endpoint=False)
    sc = SpacingComp(n_wt=n_wt, min_spacing=2, const_id="", aggregation_function=aggfunc)
    tf = Problem()
    ivc = IndepVarComp()
    ivc.add_output('x', val=np.cos(theta))
    ivc.add_output('y', val=np.sin(theta))
    tf.model.add_subsystem('ivc', ivc, promotes=['*'])
    tf.model.add_subsystem('sc', sc, promotes=['x', 'y', 'wtSeparationSquared'])
    tf.setup()
    tf.run_model()

    check = tf.check_partials(compact_print=True,
                              includes='sc*',
                              method='fd',
                              step=1e-6,
                              form='central')

    fil = {'sc': {key: val for key, val in check['sc'].items()
                  if 'penalty' not in key[0]}}

    atol = 1.e-6
    rtol = 1.e-6
    try:
        assert_check_partials(fil, atol, rtol)
    except ValueError as err:
        print(str(err))
        raise



import matplotlib.pyplot as plt
import numpy as np
import pytest

from topfarm.utils import smart_start, SmoothMax, SmoothMin, SoftMax, StrictMax, StrictMin, LogSumExpMax, LogSumExpMin
from topfarm.tests import npt
from topfarm import TopFarmProblem
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint, SpacingTypeConstraint
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent, PyWakeAEP

from py_wake.examples.data import hornsrev1
from py_wake.deficit_models.noj import NOJ
from py_wake.examples.data.iea37._iea37 import IEA37Site
from py_wake.site._site import UniformSite
from py_wake.wind_turbines import WindTurbines
from py_wake.wind_turbines.power_ct_functions import CubePowerSimpleCt


def egg_tray_map():
    x = np.arange(0, 20, 0.1)
    y = np.arange(0, 10, 0.1)
    YY, XX = np.meshgrid(y, x)
    val = np.sin(XX) + np.sin(YY)
    return XX, YY, val


def types_map():
    x = np.arange(0, 40, 0.1)
    y = np.arange(0, 20, 0.1)
    YY, XX = np.meshgrid(y, x)
    val = np.ones((4, ) + XX.shape) * (np.sin(XX) + np.sin(YY))
    val[0, :] *= 16
    val[1, :] *= 9
    val[2, :] *= 4
    return XX, YY, val


def tests_smart_start():
    xs_ref = [1.6, 14.1, 1.6, 7.9, 14.1, 7.9, 19.9, 19.9, 7.8, 5.8, 14.2,
              5.8, 1.5, 16.2, 16.2, 1.6, 3.7, 14.1, 7.9, 3.7]
    ys_ref = [1.6, 1.6, 7.9, 1.6, 7.9, 7.9, 1.6, 7.9, 5.8, 7.8, 5.8, 1.5, 5.8, 7.8, 1.5, 3.7, 1.6, 3.7, 3.7, 7.9]
    N_WT = 20
    min_space = 2.1

    XX, YY, val = egg_tray_map()
    xs, ys = smart_start(XX, YY, val, N_WT, min_space, seed=0)

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        print(np.round(xs, 1).tolist())
        print(np.round(ys, 1).tolist())

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])


def tests_smart_start_types():
    xs_ref = [26.7, 26.7, 1.6, 39.3, 1.6, 14.1, 39.3, 14.1, 33.0, 20.4, 7.9, 33.0, 20.4, 7.9, 20.4, 20.4, 20.4, 20.4, 33.0, 33.0]
    ys_ref = [1.6, 14.1, 1.6, 1.6, 14.1, 1.6, 14.1, 14.1, 7.9, 7.9, 7.9, 19.9, 19.9, 19.9, 15.7, 0.0, 3.2, 12.5, 15.8, 3.3]
    N_WT = 20
    min_space = np.array([8, 6, 4, 2]) * 1.3
    XX, YY, val = types_map()
    types = [0, 1, 2, 3]
    xs, ys, type_i = smart_start(XX, YY, val, N_WT, min_space, seed=0, types=types)
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])
    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val[0], 100)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space[type_i[i]] / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        plt.axis('equal')


def tests_smart_start_random():
    xs_ref = [14.2, 1.6, 8.1, 7.7, 14.1, 1.1, 19.8, 19.4, 3.2, 14.4,
              6.2, 16.1, 12.1, 1.1, 9.9, 16.3, 7.9, 7.9, 3.3, 12.0]
    ys_ref = [1.6, 8.2, 1.8, 7.8, 7.5, 1.6, 7.9, 2.0, 1.0, 9.6, 0.7, 8.2, 1.7, 6.1, 8.1, 1.7, 9.9, 5.7, 6.9, 7.7]

    N_WT = 20
    min_space = 2.1
    XX, YY, val = egg_tray_map()
    np.random.seed(0)

    with pytest.raises(expected_exception=AssertionError):
        xs, ys = smart_start(XX, YY, val, N_WT, min_space, random_pct=101, seed=0)
    xs, ys = smart_start(XX, YY, val, N_WT, min_space, random_pct=1, seed=0, plot=False)

    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, val, 100)
        plt.plot(XX, YY, ',k')
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        print(np.round(xs, 1).tolist())
        print(np.round(ys, 1).tolist())

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal([xs, ys], [xs_ref, ys_ref])


def tests_smart_start_no_feasible():
    XX, YY, val = egg_tray_map()
    N_WT = 20
    min_space = 5.1

    with pytest.raises(Exception, match="No feasible positions for wt 8"):
        xs, ys = smart_start(XX, YY, val, N_WT, min_space)


@pytest.mark.parametrize('seed,radius,resolution,tol', [(1, 500, 10, 5),
                                                        (0, 2000, 100, 3)])
def test_smart_start_aep_map(seed, radius, resolution, tol):
    site = IEA37Site(16)
    n_wt = 4
    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst
    wfm = NOJ(site, turbines)
    aep_comp = PyWakeAEPCostModelComponent(wfm, n_wt=n_wt)
    aep_1wt = wfm([0], [0]).aep().sum()

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep_comp,
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), radius)]
    )
    x = np.arange(-radius, radius, resolution)
    y = np.arange(-radius, radius, resolution)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep_comp.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40, plot=0, seed=seed)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(wfm(wt_x[:i], wt_y[:i]).aep().sum(['wd', 'ws']))
        aep_comp.windFarmModel(wt_x, wt_y, ws=ws_lst, wd=wd_lst).flow_map().aep_xy().plot()
        print(tf.evaluate())
        # import matplotlib.pyplot as plt
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], tol)


def test_smart_start_aep_map_types(seed=1, radius=750, resolution=10):
    x_ref = [0.0, -240.0, 700.0, -730.0, -550.0, 390.0, -600.0, 570.0, -130.0, 220.0, -180.0, 20.0, -510.0, 230.0, 670.0, 160.0]
    y_ref = [-690.0, 700.0, -70.0, -100.0, -490.0, 470.0, 450.0, -480.0, 120.0, -80.0, -300.0, 450.0, 140.0, -460.0, 300.0, 730.0]
    ts_ref = [2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 0, 0, 0, 0]
    aep_ref = 96.86027158
    site = IEA37Site(16)
    n_wt = 16
    x, y = site.initial_position.T
    wd_lst = np.arange(0, 360, 20)
    ws_lst = [10]
    types = [0, 1, 2]
    init_types = n_wt * [0]
    turbines = WindTurbines(names=['T1', 'T2', 'T3'],
                            diameters=[60, 80, 100],
                            hub_heights=[60, 80, 100],
                            powerCtFunctions=[CubePowerSimpleCt(power_rated=200 * 60 ** 2, power_unit='W'),
                                              CubePowerSimpleCt(power_rated=200 * 80 ** 2, power_unit='W'),
                                              CubePowerSimpleCt(power_rated=200 * 100 ** 2, power_unit='W')],)
    min_space = 4 * np.array([60, 80, 100])
    site.default_ws = ws_lst
    site.default_wd = wd_lst
    wfm = NOJ(site, turbines)
    aep_comp = PyWakeAEPCostModelComponent(wfm, n_wt=n_wt, additional_input=[('type', init_types)], grad_method=None)

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep_comp,
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingTypeConstraint(min_space), CircleBoundaryConstraint((0, 0), radius)],
        ext_vars={'type': init_types}
    )
    xs = np.arange(-radius, radius, resolution)
    ys = np.arange(-radius, radius, resolution)
    XX, YY = np.meshgrid(xs, ys)

    xs, ys, ts = tf.smart_start(XX, YY, aep_comp.get_aep4smart_start(wd=wd_lst, ws=ws_lst), seed=seed, types=types)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        plt.figure()
        aep_comp.windFarmModel(wt_x, wt_y, ws=ws_lst, wd=wd_lst, type=ts).flow_map().aep_xy(type=1).plot()
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal([xs, ys, ts], [x_ref, y_ref, ts_ref])
    assert tf['AEP'] >= aep_ref


def test_smart_start_aep_map_PyWakeAEP():
    site = IEA37Site(16)
    n_wt = 4

    x, y = site.initial_position[:n_wt].T
    wd_lst = np.arange(0, 360, 45)
    ws_lst = [10]
    turbines = hornsrev1.HornsrevV80()
    site = UniformSite([1], .75)
    site.default_ws = ws_lst
    site.default_wd = wd_lst

    aep = PyWakeAEP(wake_model=NOJ(site, turbines))
    aep_1wt = aep.calculate_AEP([0], [0]).sum()

    tf = TopFarmProblem(
        design_vars={'x': x, 'y': y},
        cost_comp=aep.get_TopFarm_cost_component(n_wt),
        driver=EasyScipyOptimizeDriver(),
        constraints=[SpacingConstraint(160), CircleBoundaryConstraint((0, 0), 500)]
    )
    x = np.arange(-500, 500, 10)
    y = np.arange(-500, 500, 10)
    XX, YY = np.meshgrid(x, y)

    tf.smart_start(XX, YY, aep.get_aep4smart_start(wd=wd_lst, ws=ws_lst), radius=40, seed=1)
    tf.evaluate()

    if 0:
        wt_x, wt_y = tf['x'], tf['y']
        for i, _ in enumerate(wt_x, 1):
            print(aep.calculate_AEP(wt_x[:i], wt_y[:i]).sum((1, 2)))
        X_j, Y_j, aep_map = aep.aep_map(x, y, 0, wt_x, wt_y, ws=ws_lst, wd=wd_lst)
        print(tf.evaluate())
        # import matplotlib.pyplot as plt
        c = plt.contourf(X_j, Y_j, aep_map, 100)
        plt.colorbar(c)
        plt.plot(wt_x, wt_y, '2r')
        for c in tf.model.constraint_components:
            c.plot()
        plt.axis('equal')
        plt.show()
    npt.assert_almost_equal(aep_1wt * n_wt, tf['AEP'], 5)


@pytest.mark.parametrize('max_func,scaling,ref',
                         [
                             (StrictMax(), 1, [1.0, 0.9, 0.8, 0.7, 0.6, 0.5]),
                             (StrictMin(), 1, [0, .1, .2, .3, .4, .5]),
                             (SmoothMax(.1), 1, [1.0, 0.9, 0.8, 0.69, 0.58, 0.5]),
                             (SmoothMax(.2), 1, [0.99, 0.89, 0.77, 0.65, 0.55, 0.5]),
                             (SmoothMax(100), 1000, [999.95, 899.73, 798.52, 692.81, 576.16, 500.0]),
                             (SmoothMax(50), 1000, [1000.0, 900.0, 800.0, 699.87, 596.4, 500.0]),
                             (SmoothMax(1), 1000, [1000.0, 900.0, 800.0, 700.0, 600.0, 500.0]),
                             (SmoothMin(.1), 1, [0.0, 0.1, 0.2, 0.31, 0.42, 0.5]),
                             (SmoothMin(100), 1000, [0.05, 100.27, 201.48, 307.19, 423.84, 500.0]),
                             (SmoothMin(1), 1000, [0.0, 100.0, 200.0, 300.0, 400.0, 500.0]),
                             (LogSumExpMax(.1), 1, [1.0, 0.9, 0.8, 0.7, 0.61, 0.57]),
                             (LogSumExpMax(.2), 1, [1.0, 0.9, 0.81, 0.73, 0.66, 0.64]),
                             (LogSumExpMax(200), 1000, [1001.34, 903.63, 809.72, 725.39, 662.65, 638.63]),
                             (LogSumExpMin(.1), 1, [-0.0, 0.1, 0.2, 0.3, 0.39, 0.43]),

                         ])
def test_max_funcs(max_func, scaling, ref):

    def abmax(x):
        a, b = x * scaling, (1 - x) * scaling
        return a, b, max_func([a, b], 0), max_func.gradient([a, b], 0)

    def dmax_da_fd(x):
        a, b = x * scaling, (1 - x) * scaling
        step = 1e-6
        return (max_func([a + step, b], 0) - max_func([a, b], 0)) / step

    if 0:
        x = np.arange(0, 1, .01)
        a, b, m, (dm_da, _) = abmax(x)
        plt.title(str(max_func))
        plt.plot(x, a, label='a')
        plt.plot(x, b, label='b')
        plt.plot(x, m, label='max')
        dmax_fd = dmax_da_fd(x)
        plt.plot(x, dmax_fd * scaling, label=f'fd*{scaling}')
        plt.plot(x, dm_da * scaling, '--', label=f'dmax da*{scaling}')

        x = np.arange(0, .6, .1)
        a, b, m, (dm_da, _) = abmax(x)
        plt.plot(x, m, '.')
        print(list(np.round(m, 2)))
        plt.legend()
        plt.show()

    x = np.arange(0, .6, .1)
    a, b, m, _ = abmax(x)
    npt.assert_array_almost_equal(ref, m, 2)

    x = np.arange(0, 1, .01)
    a, b, m, (dm_da, _) = abmax(x)
    if max_func.__class__.__name__.startswith('Strict'):
        mask = (x < .49) | (x > .51)
    else:
        mask = slice(None)
    npt.assert_array_almost_equal(dmax_da_fd(x)[mask], dm_da[mask], 4)



# %%
from topfarm.mongo_recorder import MongoRecorder
import matplotlib.pyplot as plt
import numpy as np
from py_wake.site.xrsite import GlobalWindAtlasSite
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition_Circle
from topfarm.constraint_components.boundary import CircleBoundaryConstraint, XYBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.cost_models.cost_model_wrappers import CostModelComponent, AEPCostModelComponent
from py_wake.examples.data.hornsrev1 import Hornsrev1Site, V80
from topfarm import TopFarmGroup, TopFarmProblem
from topfarm.plotting import XYPlotComp, NoPlot
from topfarm.easy_drivers import EasyRandomSearchDriver, EasyScipyOptimizeDriver, EasySimpleGADriver
# import os
# os.system('bash run_mongo.sh')
import subprocess

def main():
    if __name__ == '__main__':
        subprocess.Popen(['mongod'])
        plot_comp = XYPlotComp()
        # plt.rcParams['figure.dpi'] = 500

        # %% Wind turbine model and site
        wt = V80()
        D = wt.diameter()
        hub_height = wt.hub_height()
        site = Hornsrev1Site()

        # %% wind turbine coordinates (initial) and wake model
        xy = np.array([[423974, 6151447],
                       [424042, 6150891],
                       [424111, 6150335],
                       [424179, 6149779],
                       [424247, 6149224],
                       [424315, 6148668],
                       [424384, 6148112],
                       [424452, 6147556],
                       [424534, 6151447],
                       [424602, 6150891],
                       [424671, 6150335],
                       [424739, 6149779],
                       [424807, 6149224],
                       [424875, 6148668],
                       [424944, 6148112],
                       [425012, 6147556],
                       [425094, 6151447],
                       [425162, 6150891],
                       [425231, 6150335],
                       [425299, 6149779],
                       [425367, 6149224],
                       [425435, 6148668],
                       [425504, 6148112],
                       [425572, 6147556]])

        x = xy[:, 0]
        y = xy[:, 1]

        boundary = [(423500, 6.1474e6), (425700, 6.1474e6),
                    (425200, 6.1515e6), (423000, 6.1515e6)]

        wake_model = IEA37SimpleBastankhahGaussian(
            site, wt)  # select the Gaussian wake model
        simres = wake_model(xy[:, 0], xy[:, 1])
        print(simres.aep().sum())

        # %% AEP function


        def aep_func(x, y, **kwargs):
            wake_model = IEA37SimpleBastankhahGaussian(site, wt)
            simres = wake_model(x, y)
            aep = simres.aep().sum()  # AEP in GWh
            return aep


        # create an openmdao component for aep
        aep_comp = CostModelComponent(input_keys=['x', 'y'],
                                      n_wt=len(xy),
                                      cost_function=aep_func,
                                      output_key="aep",
                                      output_unit="GWh",
                                      objective=True,
                                      output_val=sum(np.zeros(len(xy))),
                                      maximize=True
                                      )

        # %% Definition of the topfarm problem
        problem = TopFarmProblem(design_vars={'x': xy[:, 0], 'y': xy[:, 1]},
                                 cost_comp=aep_comp,
                                 driver=EasyScipyOptimizeDriver(optimizer='SLSQP', maxiter=4, tol=1e-6),
                                 constraints=[SpacingConstraint(5 * D), XYBoundaryConstraint(boundary=boundary, boundary_type='polygon')],
                                 plot_comp=NoPlot(),
                                 recorder=MongoRecorder(db_name='data22', case_id='test', clean_up=True),
                                 )

        # run state before optimize
        cost, state = problem.evaluate()

        # %% OPTIMIZE
        cost, state, recorder = problem.optimize(disp=True)
        #additional_recorder = problem._additional_recorders[0]

        # %% the recorder is called in the same way as the TopfarmListRecorder

        keys = recorder.keys()
        print(keys)

        aep = recorder['aep']
        plt.plot(aep)
        plt.show()

        recorder.animate_turbineXY(duration=10, cost='aep', anim_options = {'interval': 20, 'blit': True}, filename='turbine_xy_anim')

main()


# -*- coding: utf-8 -*-
"""
Created on Fri Jan  6 14:11:37 2023

@author: mikf
"""

import numpy as np
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import CircleBoundaryConstraint
from topfarm.constraint_components.spacing import SpacingTypeConstraint
from topfarm.plotting import XYPlotComp
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.cost_models.py_wake_wrapper import PyWakeAEPCostModelComponent
from py_wake.examples.data.iea37._iea37 import IEA37Site
from py_wake.deficit_models.gaussian import IEA37SimpleBastankhahGaussian
from py_wake.wind_turbines import WindTurbines
from py_wake.wind_turbines.power_ct_functions import CubePowerSimpleCt

def main():
    if __name__ == '__main__':
        n_wt = 16
        site = IEA37Site(n_wt)
        windTurbines = WindTurbines(names=['T1', 'T2', 'T3'],
                                            diameters=[110, 130, 150],
                                            hub_heights=[110, 130, 150],
                                            powerCtFunctions = [CubePowerSimpleCt(power_rated=200 * 110 ** 2, power_unit='W'),
                                                               CubePowerSimpleCt(power_rated=200 * 130 ** 2, power_unit='W'),
                                                               CubePowerSimpleCt(power_rated=200 * 150 ** 2, power_unit='W')],)
        windFarmModel = IEA37SimpleBastankhahGaussian(site, windTurbines)
        types = 5 * [2] + 6 * [1] + 5 * [0]
        tf = TopFarmProblem(
            design_vars=dict(zip('xy', site.initial_position.T)),
            cost_comp=PyWakeAEPCostModelComponent(windFarmModel, n_wt, additional_input=[('type', np.zeros(n_wt))], grad_method=None),
            driver=EasyScipyOptimizeDriver(maxiter=50),
            constraints=[CircleBoundaryConstraint([0, 0], 1300.1),
                         SpacingTypeConstraint([windTurbines.diameter(t) * 3.5 for t in [0, 1, 2]])],
            plot_comp=XYPlotComp())
        tf['type']=types
        x = np.linspace(-1300, 1300, 101)
        y = np.linspace(-1300, 1300, 101)
        YY, XX = np.meshgrid(y, x)
        
        cost1, state1 = tf.evaluate(dict(zip('xy', site.initial_position.T)))
        cost2, state2, recorder2 = tf.optimize()
        
        tf.smart_start(XX, YY, tf.cost_comp.get_aep4smart_start(type=types))
        cost3, state3 = tf.evaluate()
        
        cost4, state4, recorder4 = tf.optimize()
        
        costs = [cost1, cost2, cost3, cost4]
        strings = ['initial', 'initial + 50 iter. optimization', 'smart start', 'smart start + 50 iter. optimization']
        for s, c in zip(strings, costs):
            print(f'{s:35}: {abs(c):.1f}')
        
main()



import numpy as np
from numpy import newaxis as na
from topfarm.constraint_components import Constraint, ConstraintComponent
import topfarm


class SpacingConstraint(Constraint):
    def __init__(self, min_spacing, units=None, aggregation_function=None, full_aggregation=False, name='spacing_comp'):
        """Initialize SpacingConstraint

        Parameters
        ----------
        min_spacing : int or float
            Minimum spacing between turbines [m]
        aggregation_function : topfarm.utils.AggregationFunction or None
            if None: compute returns all wt-wt spacings (n_wt *(n_wt-1))/2
            if AggregationFunction: compute returns an aggregated (minimum) spacing
        """
        self.min_spacing = min_spacing
        self.aggregation_function = aggregation_function
        self.full_aggregation = full_aggregation
        self.const_id = name
        self.units = units

    @property
    def constraintComponent(self):
        return self.spacing_comp

    def _setup(self, problem):
        self.n_wt = problem.n_wt
        self.spacing_comp = SpacingComp(self.n_wt, self.min_spacing, self.const_id, self.units,
                                        aggregation_function=self.aggregation_function,
                                        full_aggregation=self.full_aggregation)
        problem.model.pre_constraints.add_subsystem(self.const_id, self.spacing_comp,
                                                    promotes=[topfarm.x_key, topfarm.y_key, 'penalty_' + self.const_id, 'wtSeparationSquared'])

    def setup_as_constraint(self, problem):
        self._setup(problem)
        problem.model.add_constraint('wtSeparationSquared', lower=self.min_spacing**2)

    def setup_as_penalty(self, problem):
        self._setup(problem)


class SpacingComp(ConstraintComponent):
    """
    Calculates inter-turbine spacing for all turbine pairs.

    """

    def __init__(self, n_wt, min_spacing, const_id=None, units=None, aggregation_function=None, full_aggregation=False):
        super().__init__()
        self.n_wt = n_wt
        self.min_spacing = min_spacing
        self.const_id = const_id
        if aggregation_function:
            if full_aggregation:
                self.veclen = 1
            else:
                self.veclen = n_wt
        else:
            self.veclen = int((n_wt - 1.) * n_wt / 2.)
        self.units = units
        self.aggregation_function = aggregation_function
        self.full_aggregation = full_aggregation
        self.constraint_key = 'wtSeparationSquared'

    def setup(self):
        # Explicitly size input arrays
        self.add_input(topfarm.x_key, val=np.zeros(self.n_wt),
                       desc='x coordinates of turbines in wind dir. ref. frame', units=self.units)
        self.add_input(topfarm.y_key, val=np.zeros(self.n_wt),
                       desc='y coordinates of turbines in wind dir. ref. frame', units=self.units)
        self.add_output('penalty_' + self.const_id, val=0.0)
        # Explicitly size output array
        self.add_output(self.constraint_key, val=np.zeros(self.veclen),
                        desc='spacing of all turbines in the wind farm')

        col_pairs = np.array([(i, j) for i in range(self.n_wt - 1) for j in range(i + 1, self.n_wt)])
        if self.aggregation_function:
            self.declare_partials(self.constraint_key, [topfarm.x_key, topfarm.y_key])

            self.partial_indices = np.array([np.r_[np.where(col_pairs[:, 1] == i)[0], np.where(col_pairs[:, 0] == i)[0]]
                                             for i in range(self.n_wt)]).T
            self.partial_sign = (np.ones((self.n_wt, self.n_wt)) - 2 * np.triu(np.ones((self.n_wt, self.n_wt)), 1))[:-1]
            self.col_pairs = col_pairs
        else:
            # Sparse partial declaration
            cols = col_pairs.flatten()
            rows = np.repeat(np.arange(self.veclen), 2)

            self.declare_partials(self.constraint_key,
                                  [topfarm.x_key, topfarm.y_key],
                                  rows=rows, cols=cols)

    def compute(self, inputs, outputs):
        self.x = inputs[topfarm.x_key]
        self.y = inputs[topfarm.y_key]
        separation_squared = self._compute(self.x, self.y)
        if self.aggregation_function:
            if self.full_aggregation:
                outputs[self.constraint_key] = self.aggregation_function(separation_squared)
            else:
                outputs[self.constraint_key] = self.aggregation_function(
                    separation_squared[self.partial_indices], 0)
                # print(outputs[self.constraint_key])
        else:
            outputs[self.constraint_key] = separation_squared
        outputs['penalty_' + self.const_id] = -np.minimum(separation_squared - self.min_spacing**2, 0).sum()

    def _compute(self, x, y):
        n_wt = self.n_wt

        # compute distance matrixes
        dX, dY = [np.subtract(*np.meshgrid(xy, xy, indexing='ij')).T
                  for xy in [x, y]]
        dXY2 = dX**2 + dY**2
        # return upper triangle (above diagonal)
        return dXY2[np.triu_indices(n_wt, 1)]

    def compute_partials(self, inputs, J):
        # obtain necessary inputs
        x = inputs[topfarm.x_key]
        y = inputs[topfarm.y_key]

        # gradient of spacing [(i,j) for i=0..n_wt-1 and j=i+1..n_wt] wrt (wt_x[i], wt_x[j]) and (wt_y[i], wt_y[j])
        dS_dxij, dS_dyij = self._compute_partials(x, y)
        if self.aggregation_function:
            if self.full_aggregation:
                # gradient of aggregated (minimum) spacing wrt. spacing(i,j)
                dSagg_dS = self.aggregation_function.gradient(self._compute(x, y)).flatten()
                # partial_indices extracts the spacing elements that each wt contributes to
                # partial sign gives it the right sign
                # and finally we sum the contributions of each wt
                J[self.constraint_key, topfarm.x_key] = (
                    (dS_dxij[:, 0] * dSagg_dS)[self.partial_indices] * self.partial_sign).sum(0).T
                J[self.constraint_key, topfarm.y_key] = (
                    (dS_dyij[:, 0] * dSagg_dS)[self.partial_indices] * self.partial_sign).sum(0).T
            else:
                # gradient of aggregated (minimum) spacing wrt. spacing(i,j)
                dSdwtx = (dS_dxij[:, 0])[self.partial_indices] * self.partial_sign
                dSdwty = (dS_dyij[:, 0])[self.partial_indices] * self.partial_sign
                S = self._compute(x, y)[self.partial_indices]
                dSagg_dS = self.aggregation_function.gradient(S, 0)

                # partial_indices extracts the spacing elements that each wt contributes to
                # partial sign gives it the right sign
                # and finally we sum the contributions of each wt

                dSagg_dwtx = dSdwtx * dSagg_dS * self.partial_sign
                dSagg_dwty = dSdwty * dSagg_dS * self.partial_sign

                dSagg_dx = np.zeros((self.n_wt, self.n_wt))  # np.diag((dSagg_dwtx).sum(0))
                dSagg_dy = np.zeros((self.n_wt, self.n_wt))  # np.diag((dSagg_dwty).sum(0))
                i = range(self.n_wt)
                for j in range(dSagg_dwtx.shape[0]):
                    ai, bi = self.col_pairs[self.partial_indices][j, i, :].T
                    dSagg_dx[ai, i] += dSagg_dwtx[j, i]
                    dSagg_dx[bi, i] -= dSagg_dwtx[j, i]
                    dSagg_dy[ai, i] += dSagg_dwty[j, i]
                    dSagg_dy[bi, i] -= dSagg_dwty[j, i]

                J[self.constraint_key, topfarm.x_key] = dSagg_dx.T
                J[self.constraint_key, topfarm.y_key] = dSagg_dy.T
        else:
            # populate Jacobian dict
            J[self.constraint_key, topfarm.x_key] = dS_dxij.flatten()
            J[self.constraint_key, topfarm.y_key] = dS_dyij.flatten()

    def _compute_partials(self, x, y):
        # get number of turbines
        n_wt = self.n_wt

        # S = (xi-xj)^2 + (yi-yj)^2 = dx^2 + dy^2
        # dS/dx = [dS/dx_i, dS/dx_j]  = -2dx, 2dx

        # compute distance matrixes
        dX, dY = [np.subtract(*np.meshgrid(xy, xy, indexing='ij')).T
                  for xy in [x, y]]
        # upper triangle -> 1 row per WT pair [(0,1), (0,2),..(n-1,n)]
        dx, dy = dX[np.triu_indices(n_wt, 1)], dY[np.triu_indices(n_wt, 1)]

        dSdx = np.array([-2 * dx, 2 * dx]).T
        dSdy = np.array([-2 * dy, 2 * dy]).T
        return dSdx, dSdy

    def plot(self, ax=None):
        from matplotlib.pyplot import Circle
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()

        def get_xy(xy):
            if not hasattr(self, xy):
                setattr(self, xy, dict(self.list_inputs(out_stream=None))[f'pre_constraints.{self.name}.{xy}']['value'])
            xy = getattr(self, xy)
            return xy if not isinstance(xy, tuple) else xy[0]

        for x, y in zip(get_xy('x'), get_xy('y')):
            circle = Circle((x, y), self.min_spacing / 2, color='k', ls='--', fill=False)
            ax.add_artist(circle)

    def satisfy(self, state, n_iter=100, step_size=0.1):
        x, y = [state[xy].astype(float) for xy in [topfarm.x_key, topfarm.y_key]]
        pair_i, pair_j = np.triu_indices(len(x), 1)
        for _ in range(n_iter):
            dist = self._compute(x, y)
            dx, dy = self._compute_partials(x, y)
            index = np.argmin(dist)

            if dist[index] < self.min_spacing**2:
                i, j = pair_i[index], pair_j[index]
                x[i] += dx[index, 0] * step_size
                x[j] += dx[index, 1] * step_size
                y[i] += dy[index, 0] * step_size
                y[j] += dy[index, 1] * step_size
            else:
                break
        state.update({topfarm.x_key: x, topfarm.y_key: y})
        return state


class SpacingTypeConstraint(SpacingConstraint):
    def __init__(self, min_spacing, units=None, aggregation_function=None, full_aggregation=False, name='spacing_type_comp'):
        """Initialize SpacingConstraint

        Parameters
        ----------
        min_spacing : array_like
            Minimum spacing around turbines [m] (diameter of circle around a turbine that no other turbines can occupy)
        aggregation_function : topfarm.utils.AggregationFunction or None
            if None: compute returns all wt-wt spacings (n_wt *(n_wt-1))/2
            if AggregationFunction: compute returns an aggregated (minimum) spacing
        """
        super().__init__(min_spacing=min_spacing, units=units, aggregation_function=aggregation_function,
                         full_aggregation=full_aggregation, name=name)
        self.min_spacing = np.asarray(min_spacing)

    def _setup(self, problem):
        self.n_wt = problem.n_wt
        self.spacing_comp = SpacingTypeComp(self.n_wt, self.min_spacing, self.const_id, self.units,
                                            aggregation_function=self.aggregation_function,
                                            full_aggregation=self.full_aggregation)
        problem.model.pre_constraints.add_subsystem(self.const_id, self.spacing_comp,
                                                    promotes=[topfarm.x_key, topfarm.y_key, topfarm.type_key,
                                                              'penalty_' + self.const_id, 'wtRelativeSeparationSquared'])

    def setup_as_constraint(self, problem):
        self._setup(problem)
        problem.model.add_constraint('wtRelativeSeparationSquared', lower=0)


class SpacingTypeComp(SpacingComp):
    """
    Calculates inter-turbine spacing for all turbine pairs.

    """

    def __init__(self, n_wt, min_spacing, const_id=None, units=None, aggregation_function=None, full_aggregation=False, types=None):
        super().__init__(n_wt=n_wt, min_spacing=min_spacing, const_id=const_id, units=units,
                         aggregation_function=aggregation_function, full_aggregation=full_aggregation)
        self.constraint_key = 'wtRelativeSeparationSquared'
        self.types = types

    def setup(self):
        super().setup()
        self.add_input(topfarm.type_key, val=self.types or np.zeros(self.n_wt),
                       desc='turbine type number')

    def compute(self, inputs, outputs):
        self.x = inputs[topfarm.x_key]
        self.y = inputs[topfarm.y_key]
        self.type = inputs[topfarm.type_key]
        relative_separation_squared = self._compute(self.x, self.y, self.type)
        if self.aggregation_function:
            if self.full_aggregation:
                outputs[self.constraint_key] = self.aggregation_function(relative_separation_squared)
            else:
                outputs[self.constraint_key] = self.aggregation_function(
                    relative_separation_squared[self.partial_indices], 0)
                # print(outputs[self.constraint_key])
        else:
            outputs[self.constraint_key] = relative_separation_squared
        outputs['penalty_' + self.const_id] = -np.minimum(relative_separation_squared, 0).sum()

    def get_min_eff_spacing(self, t):
        return (self.min_spacing[np.atleast_1d(t).astype(int)][:, na] + self.min_spacing[np.atleast_1d(t).astype(int)][na, :]) / 2

    def _compute(self, x, y, t):
        n_wt = self.n_wt
        # compute distance matrixes
        dX, dY = [np.subtract(*np.meshgrid(xy, xy, indexing='ij')).T
                  for xy in [x, y]]
        dXY2 = dX**2 + dY**2 - self.get_min_eff_spacing(t)**2
        # return upper triangle (above diagonal)
        return dXY2[np.triu_indices(n_wt, 1)]

    def plot(self, ax=None):
        from matplotlib.pyplot import Circle
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()

        def get_xy(xy):
            if not hasattr(self, xy):
                setattr(self, xy, dict(self.list_inputs(out_stream=None))[f'pre_constraints.{self.name}.{xy}']['value'])
            xy = getattr(self, xy)
            return xy if not isinstance(xy, tuple) else xy[0]

        for x, y, t in zip(get_xy('x'), get_xy('y'), get_xy('type')):
            circle = Circle((x, y), self.get_min_eff_spacing(t).ravel() / 2, color='k', ls='--', fill=False)
            ax.add_artist(circle)

    def satisfy(self, state, n_iter=100, step_size=0.1):
        x, y, t = [state[xy].astype(float) for xy in [topfarm.x_key, topfarm.y_key, topfarm.type_key]]
        pair_i, pair_j = np.triu_indices(len(x), 1)
        for _ in range(n_iter):
            dist = self._compute(x, y, t)
            dx, dy = self._compute_partials(x, y)
            index = np.argmin(dist)

            if dist[index] < self.get_min_eff_spacing(t)[np.triu_indices(self.n_wt, 1)][index]**2:
                i, j = pair_i[index], pair_j[index]
                x[i] += dx[index, 0] * step_size
                x[j] += dx[index, 1] * step_size
                y[i] += dy[index, 0] * step_size
                y[j] += dy[index, 1] * step_size
            else:
                break
        state.update({topfarm.x_key: x, topfarm.y_key: y})
        return state



import numpy as np
import matplotlib.pyplot as plt

from py_wake.examples.data.iea37._iea37 import IEA37_WindTurbines
from py_wake.deficit_models.gaussian import BastankhahGaussian
from py_wake.examples.data.lillgrund import LillgrundSite
from py_wake.utils.gradients import autograd
from py_wake.examples.data.hornsrev1 import Hornsrev1Site, HornsrevV80
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.easy_drivers import EasySGDDriver, EasyScipyOptimizeDriver
from topfarm.plotting import AggregatedConstraintsPlotComponent, XYPlotComp
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.recorders import TopFarmListRecorder


def main():
    if __name__ == '__main__':
        plt.close('all')

        site = LillgrundSite()
        site.interp_method = 'linear'
        windTurbines = HornsrevV80() 
        wake_model = BastankhahGaussian(site, windTurbines) 

        x_rows = 5
        y_rows = 5
        spacing = 5
        xu, yu = (x_rows * spacing * windTurbines.diameter(), y_rows * spacing * windTurbines.diameter())
        #x = np.array([np.linspace(0, xu, x_rows) for _ in range(y_rows)]).flatten()
        #y = np.array([np.ones(x_rows) * spacing * windTurbines.diameter() * ii for ii in range(y_rows)]).flatten()
        np.random.seed(1)
        x = np.random.uniform(0, xu, x_rows * y_rows)
        y = np.random.uniform(0, yu, x_rows * y_rows)
        plt.figure()
        plt.plot(x, y, '2k')
        dirs = np.arange(0, 360, 1)
        freqs = site.local_wind(x, y).Sector_frequency_ilk[0, :, 0]
        As = site.local_wind(x, y).Weibull_A_ilk[0, :, 0]
        ks = site.local_wind(x, y).Weibull_k_ilk[0, :, 0]
        samps = 50
        boundary = np.array([(0, 0), (xu, 0), (xu, yu), (0, yu)])
        constraint_comp = XYBoundaryConstraint(boundary, 'rectangle')

        random = True
        # T = 2

        #wd = np.arange(0, 360, 20)
        #ws = np.arange(3, 25, 1)
        def aep_func(x, y, full=False, **kwargs):
            if not random:
                np.random.seed(0)
            idx = np.random.choice(np.arange(dirs.size), samps, p=freqs)
            wd = dirs[idx]
            A = As[idx]
            k = ks[idx]
            if not random:
                np.random.seed(0)
            ws = A * np.random.weibull(k)
            if full:
               wd = np.arange(0, 360, 1)
               ws = np.arange(3, 25, 1)
            return wake_model(x, y, wd=wd, ws=ws, time=not full).aep().sum().values * 1e6

        def aep_func2(x, y, **kwargs):
            wd = np.arange(0, 360, 1)
            ws = np.arange(3, 25, 1)
            return wake_model(x, y, wd=wd, ws=ws).aep().sum().values * 1e6

        def aep_jac(x, y, **kwargs):
            if not random:
                np.random.seed(0)
            idx = np.random.choice(np.arange(dirs.size), samps, p=freqs)
            wd = dirs[idx]
            A = As[idx]
            k = ks[idx]
            if not random:
                np.random.seed(0)
            ws = A * np.random.weibull(k)
            jx, jy = wake_model.aep_gradients(gradient_method=autograd, wrt_arg=['x', 'y'], x=x, y=y, ws=ws, wd=wd, time=True)
            return np.array([np.atleast_2d(jx), np.atleast_2d(jy)]) * 1e6

        def aep_jac2(x, y, **kwargs):
            wd = np.arange(0, 360, 1)
            ws = np.arange(3, 25, 1)
            jx, jy = wake_model.aep_gradients(gradient_method=autograd, wrt_arg=['x', 'y'], x=x, y=y, ws=ws, wd=wd, time=False)
            return np.array([np.atleast_2d(jx), np.atleast_2d(jy)]) * 1e6

        n_wt=x.size
        aep_comp = CostModelComponent(input_keys=['x','y'], n_wt=n_wt, cost_function=aep_func, objective=True, cost_gradient_function=aep_jac, maximize=True)
        aep_comp2 = CostModelComponent(input_keys=['x','y'], n_wt=n_wt, cost_function=aep_func2, objective=True, cost_gradient_function=aep_jac2, maximize=True)
        cost_comps = [aep_comp2, aep_comp]

        def constr_aggr_func(wtSeparationSquared, boundaryDistances, **kwargs):
            separation_constraint = wtSeparationSquared - (2 * windTurbines.diameter()) ** 2
            separation_constraint = separation_constraint[separation_constraint < 0]
            distance_constraint = boundaryDistances
            distance_constraint = distance_constraint[distance_constraint < 0]
            # return np.sum(-1 * np.minimum(wtSeparationSquared - (2 * windTurbines.diameter()) ** 2, 0)) + np.sum(np.minimum(boundaryDistances, 0) ** 2)
            return np.sum(-1 * separation_constraint) + np.sum(distance_constraint ** 2)

        def constr_aggr_grad(wtSeparationSquared, boundaryDistances, **kwargs):
            separation_constraint = wtSeparationSquared - (2 * windTurbines.diameter()) ** 2
            J_separation = np.zeros_like(wtSeparationSquared)
            J_separation[np.where(separation_constraint < 0)] = -1
            J_distance = np.zeros_like(boundaryDistances)
            J_distance[np.where(boundaryDistances < 0)] = 2 * boundaryDistances[np.where(boundaryDistances < 0)]
            return [[J_separation], [J_distance]]

        name = 'sgd_constraint'
        component_args = {'input_keys': [('wtSeparationSquared', np.zeros(int(n_wt * (n_wt - 1) / 2))),
                                         ('boundaryDistances', np.zeros((n_wt, 4)))],
                          'n_wt': n_wt,
                          'cost_function': constr_aggr_func,
                          'cost_gradient_function': constr_aggr_grad,
                          'objective': False,
                          'output_keys': [(name, 0)],
                          'use_penalty': False}
        constraint_args = {'name': name, 'lower': 0}
        from topfarm.constraint_components.constraint_aggregation import ConstraintAggregation

        driver_names = ['SLSQP', 'SGD']
        drivers = [EasyScipyOptimizeDriver(maxiter=200, tol=1e-3), EasySGDDriver(maxiter=400, learning_rate=windTurbines.diameter() / 5, max_time=30 * 60, speedupSGD=True, sgd_thresh=0.5)]
        driver_no = 1
        constraints = [[SpacingConstraint(2 * windTurbines.diameter()), constraint_comp], ConstraintAggregation([SpacingConstraint(2 * windTurbines.diameter()), constraint_comp],
                                                  component_args=component_args, constraint_args=constraint_args)]
        ecs = [1e2, 1]
        tf = TopFarmProblem(
                design_vars={'x':(x,0,xu), 'y':(y,0,yu)}, # setting up our two turbines as design variables
                cost_comp=cost_comps[driver_no], # using dummy cost model
                constraints=constraints[driver_no], # constraint set up for the boundary type provided)
                driver=drivers[driver_no],
                plot_comp=[XYPlotComp(), AggregatedConstraintsPlotComponent()][driver_no],
                expected_cost=ecs[driver_no],
                )
        #tf.driver.learning_rate = windTurbines.diameter() / 5
        if 1:
            cost, state, recorder = tf.optimize()
            recorder.save(f'{driver_names[driver_no]}')

        if 0:
            #plt.figure()
            plt.clf()
            for i in range(2):
                rec = TopFarmListRecorder().load(f'recordings/{driver_names[i]}')
                if driver_names[i] == 'SGD':
                    aep = []
                    for x, y in zip(rec['x'], rec['y']):
                        aep.append(aep_func2(x, y))
                else:
                    aep = rec['Cost']
                    # else:
                    #     aep = 
                plt.plot(rec['timestamp']-rec['timestamp'][0], aep, label=driver_names[i])
            plt.legend()
            plt.savefig('conv')
            plt.clf()
            # plt.plot()

main()



"""
Class definition for MongoRecorder, which provides dictionary inspired by SQLite.
"""

from copy import deepcopy
from collections import OrderedDict
import os
from itertools import chain
import json
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from matplotlib import animation
# from six import iteritems
from openmdao.recorders.case_recorder import CaseRecorder
from openmdao.core.driver import Driver
from openmdao.core.system import System
from openmdao.core.problem import Problem
from openmdao.solvers.solver import Solver
from pymongo import MongoClient
from topfarm.drivers.random_search_driver import RandomSearchDriver
import random
from datetime import datetime
# import pandas as pd
# separator, cannot be a legal char for names
META_KEY_SEP = '!'

_container_classes = (list, tuple, set)


def make_serializable(o):
    """
    Recursively convert numpy types to native types for JSON serialization.
    Parameters
    ----------
    o : object
        the object to be converted
    Returns
    -------
    object
        The converted object.
    """
    if isinstance(o, _container_classes):
        return [make_serializable(item) for item in o]
    elif isinstance(o, np.number):
        return o.item()
    elif isinstance(o, np.ndarray):
        return make_serializable(o.tolist())
    elif hasattr(o, '__dict__'):
        return make_serializable(o.__class__.__name__)
    else:
        return clean_keys(o)


def clean_str(k):
    """
    Function to replace '.' with '%'.
    Used when writing to Mongo database.
    Helper function for clean_keys() function.

    Parameters
    ----------
    k : str

    Returns
    -------
    updated k with replaced "." for "%"
    """
    return k.replace('.', '%')


def clean_keys(o):
    """
    Function removes extracts keys from layered dict.
    Helper function for make_serializable() function.

    Parameters
    ----------
    o : object
        If objects is a dict, flattens.

    Returns
    -------
    Updated o : object
        Flattened dict where applicable
    """
    if isinstance(o, dict):
        new_o = {}
        for k, v in o.items():
            new_o[clean_str(k)] = v
        return new_o
    else:
        return o


class RemoteMongo():
    def __init__(self, uri=None, host=None, port=27017, ip='localhost', password=None, user=None, db_name='database', uri_type=None, **kwargs):
        self.host = host
        self.port = port
        self.ip = ip
        self.uri = uri
        self.user = user
        self.password = password
        self.db_name = db_name
        self.uri_type = uri_type
        self.kwargs = kwargs

    def open(self):
        if self.uri:
            uri = self.uri
            self.kwargs.update({'host': uri})
        elif self.host:
            if not self.uri_type:
                uri = f"mongodb://{self.user}:{self.password}@{self.host}"
            elif self.uri_type == 'srv':
                uri = f"mongodb+srv://{self.user}:{self.password}@{self.host}"
        else:
            self.kwargs.update({'host': f'mongodb://{self.ip}:{self.port}',
                                'username': self.user,
                                'password': self.password,
                                'authSource': 'database_str'})

        self.client = MongoClient(**self.kwargs)
        self.db = self.client[self.db_name]
        return self.client

    def __enter__(self):
        self.open()
        return self.db

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.client.close()


class MongoRecorder(CaseRecorder):
    """
    Recorder that saves cases in Mongo DB.

    Attributes
    ----------
    _record_viewer_data : bool
        Flag indicating whether to record data needed to generate N2 diagram.
    connection :  connection object
        Connection to the database.
    _abs2prom : {'input': dict, 'output': dict}
        Dictionary mapping absolute names to promoted names.
    _prom2abs : {'input': dict, 'output': dict}
        Dictionary mapping promoted names to absolute names.
    _abs2meta : {'name': {}}
        Dictionary mapping absolute variable names to their metadata including units,
        bounds, and scaling.
    _database_initialized : bool
        Flag indicating whether or not the database has been initialized.
    _record_on_proc : bool
        Flag indicating whether to record on this processor when running in parallel.
    """

    def __init__(self, uri=None, host=None, ip='localhost', port=27017, user=None, password=None, uri_type=None, record_viewer_data=True, ssh_tunnel=False,
                 db_name='database', case_id=None, clean_up=False, with_mpi=False, client_args={}):
        """
        Initialize the MongoRecorder.

        Parameters
        ----------
        ip: str
            ip of the mongo server
        port: int
            port of the mongo server
        record_viewer_data : bool, optional
            If True, record data needed for visualization.
        """
        self._ip = ip
        self._port = port
        self._host = host
        self._uri = uri
        self._user = user
        self._password = password
        self._uri_type = uri_type
        self._client_args = client_args
        self.case_id = case_id
        self.db_name = db_name
        self.ssh_tunnel = ssh_tunnel
        self._record_viewer_data = record_viewer_data
        self._record_metadata = True
        self.metadata_connection = None
        self._started = set()

        self._abs2prom = {'input': {}, 'output': {}}
        self._prom2abs = {'input': {}, 'output': {}}
        self._abs2meta = {}
        self._database_initialized = False

        # default to record on all procs when running in parallel
        self._record_on_proc = True
        self.lastrowid = 0

        self.connection = True

        # testing
        self.driver_iteration_dict = {}

        # Generating a random id
        random.seed(datetime.now())
        self.run_id = int(random.random() * 10000000)

        if with_mpi:
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            self._parallel = False
            self._record_on_proc = True
            self.rank = comm.rank
            self.size = comm.size
        else:
            self.rank = -1
            self.size = 1
            self._parallel = False

        if clean_up:
            with self.mongodb() as db:
                if with_mpi:
                    if comm.rank == 0:
                        delete = True
                    else:
                        delete = False
                else:
                    delete = True

                if delete:
                    db.driver_iterations.delete_many({'case_id': case_id})
                    db.solver_iterations.delete_many({'case_id': case_id})
                    db.system_iterations.delete_many({'case_id': case_id})
                    db.problem_cases.delete_many({'case_id': case_id})
                    db.global_iterations.delete_many({'case_id': case_id})
                    db.metadata.delete_many({'case_id': case_id})
                    db.driver_derivatives.delete_many({'case_id': case_id})
                    db.solver_metadata.delete_many({'case_id': case_id})
                    db.system_metadata.delete_many({'case_id': case_id})
                    db.driver_metadata.delete_many({'case_id': case_id})

        super(MongoRecorder, self).__init__(record_viewer_data)
        if with_mpi:
            self._parallel = False
            self._record_on_proc = True
            self.rank = comm.rank
            self.size = comm.size
        else:
            self.rank = -1
            self.size = 1
            self._parallel = False

        # testing
        self.iteration_coordinate_lst = []

    def get(self, key):
        with self.mongodb() as db:
            res = list(db.driver_iterations.find({'case_id': self.case_id}, {key: 1}))
        return np.asarray([x[key] for x in res])

    def __getitem__(self, key):
        return self.get(key)

    def keys(self):
        with self.mongodb() as db:
            return list(db.driver_iterations.find_one())

    def _initialize_database(self):
        """
        Initialize the database.
        """
        # TODO

        self._database_initialized = True

    def mongodb(self):
        if self.ssh_tunnel:
            print('Warning: SSH database access not available in this Topfarm version!')
            # return EC_Mongo()
        else:
            client = RemoteMongo(uri=self._uri, host=self._host, port=self._port, ip=self._ip, user=self._user, password=self._password, db_name=self.db_name, uri_type=self._uri_type, **self._client_args)
            return client

    def exists(self):
        with self.mongodb() as db:
            return db.driver_iterations.find_one({'case_id': self.case_id}) is not None

    @property
    def num_cases(self):
        return len(self.iteration_coordinate_lst)

    def _cleanup_abs2meta(self):
        """
        Convert all abs2meta variable properties to a form that can be dumped as JSON.
        """
        for name in self._abs2meta:
            for prop in self._abs2meta[name]:
                self._abs2meta[name][prop] = make_serializable(self._abs2meta[name][prop])

    def _cleanup_var_settings(self, var_settings):
        """
        Convert all var_settings variable properties to a form that can be dumped as JSON.

        Parameters
        ----------
        var_settings : dict
            Dictionary mapping absolute variable names to variable settings.

        Returns
        -------
        var_settings : dict
            Dictionary mapping absolute variable names to var settings that are JSON compatible.
        """
        # otherwise we trample on values that are used elsewhere
        var_settings = deepcopy(var_settings)
        for name in var_settings:
            for prop in var_settings[name]:
                var_settings[name][prop] = make_serializable(var_settings[name][prop])
        return var_settings

    def startup(self, recording_requester, comm=None):
        """
        Prepare for a new run and create/update the abs2prom and prom2abs variables.

        Parameters
        ----------
        recording_requester : object
            Object to which this recorder is attached.
        """

        # we only want to set up recording once for each recording_requester
        if recording_requester in self._started:
            return

        super(MongoRecorder, self).startup(recording_requester)

        if not self._database_initialized:
            self._initialize_database()

        driver = None

        # grab the system and driver
        if isinstance(recording_requester, Driver):
            system = recording_requester._problem().model
            driver = recording_requester
        elif isinstance(recording_requester, System):
            system = recording_requester
            driver = None
        elif isinstance(recording_requester, Problem):
            system = recording_requester.model
            driver = recording_requester.driver
        elif isinstance(recording_requester, Solver):
            system = recording_requester._system()
            driver = None
        else:
            raise ValueError('Driver encountered a recording_requester it cannot handle'
                             ': {0}'.format(recording_requester))

        states = system._list_states_allprocs()

        if self.connection:

            if driver is None:
                desvars = system.get_design_vars(True, get_sizes=False, use_prom_ivc=False)
                responses = system.get_responses(True, get_sizes=False)
                constraints = OrderedDict()
                objectives = OrderedDict()
                for name, data in responses.items():
                    if data['type'] == 'con':
                        constraints[name] = data
                    else:
                        objectives[name] = data
            else:
                desvars = driver._designvars.copy()
                responses = driver._responses.copy()
                constraints = driver._cons.copy()
                objectives = driver._objs.copy()

            inputs = list(system.abs_name_iter('input', local=False, discrete=True))
            outputs = list(system.abs_name_iter('output', local=False, discrete=True))

            # var_order = system._get_vars_exec_order(inputs=True, outputs=True)

            # merge current abs2prom and prom2abs with this system's version
            self._abs2prom['input'].update(system._var_allprocs_abs2prom['input'])
            self._abs2prom['output'].update(system._var_allprocs_abs2prom['output'])
            for v, abs_names in system._var_allprocs_prom2abs_list['input'].items():
                if v not in self._prom2abs['input']:
                    self._prom2abs['input'][v] = abs_names
                else:
                    self._prom2abs['input'][v] = list(set(chain(self._prom2abs['input'][v],
                                                                abs_names)))

            # for outputs, there can be only one abs name per promoted name
            for v, abs_names in system._var_allprocs_prom2abs_list['output'].items():
                self._prom2abs['output'][v] = abs_names

            # absolute pathname to metadata mappings for continuous & discrete variables
            # discrete mapping is sub-keyed on 'output' & 'input'
            real_meta_in = system._var_allprocs_abs2meta['input']
            real_meta_out = system._var_allprocs_abs2meta['output']
            disc_meta_in = system._var_allprocs_discrete['input']
            disc_meta_out = system._var_allprocs_discrete['output']

            full_var_set = [(outputs, 'output'),
                            (desvars, 'desvar'), (responses, 'response'),
                            (objectives, 'objective'), (constraints, 'constraint')]

            for var_set, var_type in full_var_set:
                for name in var_set:

                    # Design variables, constraints and objectives can be requested by input name.
                    if var_type != 'output':
                        try:
                            name = var_set[name]['ivc_source']
                        except KeyError:
                            name = var_set[name]['source']

                    if name not in self._abs2meta:
                        try:
                            self._abs2meta[name] = real_meta_out[name].copy()
                        except KeyError:
                            self._abs2meta[name] = disc_meta_out[name].copy()
                        self._abs2meta[name]['type'] = []
                        self._abs2meta[name]['explicit'] = name not in states

                    if var_type not in self._abs2meta[name]['type']:
                        self._abs2meta[name]['type'].append(var_type)

            for name in inputs:
                try:
                    self._abs2meta[name] = real_meta_in[name].copy()
                except KeyError:
                    self._abs2meta[name] = disc_meta_in[name].copy()
                self._abs2meta[name]['type'] = ['input']
                self._abs2meta[name]['explicit'] = True

            # merge current abs2meta with this system's version
            for name, meta in self._abs2meta.items():
                for io in ('input', 'output'):
                    if name in system._var_allprocs_abs2meta[io]:
                        meta.update(system._var_allprocs_abs2meta[io][name])
                        break

        self._cleanup_abs2meta()

        # store the updated abs2prom and prom2abs
        abs2prom = json.dumps(self._abs2prom)
        prom2abs = json.dumps(self._prom2abs)
        abs2meta = json.dumps(self._abs2meta)

        var_settings = {}
        var_settings.update(desvars)
        var_settings.update(objectives)
        var_settings.update(constraints)
        var_settings = self._cleanup_var_settings(var_settings)
        var_settings_json = json.dumps(var_settings)

        with self.mongodb() as db:
            data = {
                'case_id': self.case_id,
                'abs2prom': abs2prom,
                'prom2abs': prom2abs,
                'abs2meta': abs2meta,
                'var_settings_json': var_settings_json}
            db.metadata.insert_one(data)
        self._started.add(recording_requester)

    def record_iteration(self, recording_requester, data, metadata, **kwargs):
        """
        Route the record_iteration call to the proper method.

        Parameters
        ----------
        recording_requester : object
            System, Solver, Driver in need of recording.
        metadata : dict, optional
            Dictionary containing execution metadata.
        data : dict
            Dictionary containing desvars, objectives, constraints, responses, and System vars.
        **kwargs : keyword args
            Some implementations of record_iteration need additional args.
        """
        self._counter += 1

        self._iteration_coordinate = \
            recording_requester._recording_iter.get_formatted_iteration_coordinate()

        if isinstance(recording_requester, Driver):
            self.record_iteration_driver(recording_requester, data, metadata)
        elif isinstance(recording_requester, System):
            self.record_iteration_system(recording_requester, data, metadata)
        elif isinstance(recording_requester, Solver):
            self.record_iteration_solver(recording_requester, data, metadata)
        elif isinstance(recording_requester, Problem):
            self.record_iteration_problem(recording_requester, data, metadata)
        else:
            raise ValueError("Recorders must be attached to Drivers, Systems, or Solvers.")

    def record_iteration_driver(self, driver, data, metadata):
        """
        Record data and metadata from a Driver.

        Parameters
        ----------
        driver : Driver
            Driver in need of recording.
        data : dict
            Dictionary containing desvars, objectives, constraints, responses, and System vars.
        metadata : dict
            Dictionary containing execution metadata.
        """
        if not self._database_initialized:
            raise RuntimeError(f"{driver.msginfo} attempted to record iteration to "
                               f"'{self._filepath}', but database is not initialized;"
                               " `run_model()`, `run_driver()`, or `final_setup()` "
                               "must be called after adding a recorder.")

        if self.connection:
            outputs = data['output']
            outputs = {self._abs2prom['output'][k]: v for k, v in outputs.items()}
            inputs = data['input']
            inputs = {self._abs2prom['input'][k]: v for k, v in inputs.items()}
            residuals = data['residual']

            # convert to list so this can be dumped as JSON
            for in_out_resid in (inputs, outputs, residuals):
                if in_out_resid is None:
                    continue
                for var in in_out_resid:
                    in_out_resid[var] = make_serializable(in_out_resid[var])

            # outputs_text = json.dumps(outputs)
            # inputs_text = json.dumps(inputs)
            # residuals_text = json.dumps(residuals)

            if isinstance(driver, RandomSearchDriver):
                max_step = driver.randomize_func.max_step
            else:
                max_step = None

            with self.mongodb() as db:
                self.lastrowid += 1
                data = {
                    'case_id': self.case_id,
                    'run_id': self.run_id,
                    'counter': self._counter,
                    'timestamp': metadata['timestamp'],
                    'success': metadata['success'],
                    'msg': metadata['msg'],
                    'rank': self.rank,
                    'size': self.size,
                    'max_step': max_step
                }
                data.update(make_serializable(inputs))
                data.update(make_serializable(outputs))
                db.driver_iterations.insert_one(data)

                data = {
                    'case_id': self.case_id,
                    'run_id': self.run_id,
                    'counter': self._counter,
                    'record_type': 'driver',
                    'rowid': self.lastrowid,
                    'source': driver._get_name()
                }
                db.global_iterations.insert_one(data)

    def record_iteration_problem(self, data, metadata):
        """
        Record data and metadata from a Problem.

        Parameters
        ----------
        recording_requester : object
            Problem in need of recording.
        data : dict
            Dictionary containing desvars, objectives, and constraints.
        metadata : dict
            Dictionary containing execution metadata.
        """
        pass
        # outputs = data['output']

        # # convert to list so this can be dumped as JSON
        # if outputs is not None:
        #     for var in outputs:
        #         outputs[var] = make_serializable(outputs[var])

        # with self.mongodb() as db:
        #     data = {
        #         'case_id': self.case_id,
        #         'counter': self._counter,
        #         'timestamp': metadata['timestamp'],
        #         'success': metadata['success'],
        #         'msg': metadata['msg'],
        #     }
        #     data.update(make_serializable(outputs))
        #     db.problem_cases.inser_one(data)

    def record_iteration_system(self, recording_requester, data, metadata):
        """
        Record data and metadata from a System.

        Parameters
        ----------
        recording_requester : System
            System in need of recording.
        data : dict
            Dictionary containing inputs, outputs, and residuals.
        metadata : dict
            Dictionary containing execution metadata.
        """
        pass
        # inputs = data['i']
        # outputs = data['o']
        # residuals = data['r']

        # # convert to list so this can be dumped as JSON
        # for i_o_r in (inputs, outputs, residuals):
        #     if i_o_r is None:
        #         continue
        #     for var in i_o_r:
        #         i_o_r[var] = make_serializable(i_o_r[var])

        # with self.mongodb() as db:
        #     data = {
        #         'case_id': self.case_id,
        #         'iteration_coordinate': self._iteration_coordinate,
        #         'counter': self._counter,
        #         'timestamp': metadata['timestamp'],
        #         'success': metadata['success'],
        #         'msg': metadata['msg'],
        #         'residuals': make_serializable(residuals)
        #     }
        #     data.update(make_serializable(inputs))
        #     data.update(make_serializable(outputs))
        #     db.system_iterations.inser_one(data)

        #     # get the pathname of the source system
        #     source_system = recording_requester.pathname
        #     if source_system == '':
        #         source_system = 'root'

        #     data = {
        #         'case_id': self.case_id,
        #         'counter': self._counter,
        #         'record_type': 'system',
        #         'rowid': self.lastrowid,
        #         'source': source_system
        #     }
        #     db.global_iterations.insert_one(data)

    def record_iteration_solver(self, recording_requester, data, metadata):
        """
        Record data and metadata from a Solver.

        Parameters
        ----------
        recording_requester : Solver
            Solver in need of recording.
        data : dict
            Dictionary containing outputs, residuals, and errors.
        metadata : dict
            Dictionary containing execution metadata.
        """
        pass
        # abs = data['abs']
        # rel = data['rel']
        # inputs = data['i']
        # outputs = data['o']
        # residuals = data['r']

        # with self.mongodb() as db:
        #     data = {
        #         'case_id': self.case_id,
        #         'iteration_coordinate': self._iteration_coordinate,
        #         'counter': self._counter,
        #         'timestamp': metadata['timestamp'],
        #         'success': metadata['success'],
        #         'msg': metadata['msg'],
        #         'abs_err': abs,
        #         'rel_err': rel,
        #         'residuals': make_serializable(residuals)
        #     }
        #     data.update(make_serializable(inputs))
        #     data.update(make_serializable(outputs))
        #     db.solver_iterations.inser_one(data)

        #     # get the pathname of the source system
        #     source_system = recording_requester._system.pathname
        #     if source_system == '':
        #         source_system = 'root'

        #     solver_type = recording_requester.SOLVER[0:2]
        #     if solver_type == 'NL':
        #         source_solver = source_system + '.nonlinear_solver'
        #     elif solver_type == 'LS':
        #         source_solver = source_system + '.nonlinear_solver.linesearch'
        #     else:
        #         raise RuntimeError("Solver type '%s' not recognized during recording. "
        #                            "Expecting NL or LS" % recording_requester.SOLVER)

        #     data = {
        #         'case_id': self.case_id,
        #         'counter': self._counter,
        #         'record_type': 'solver',
        #         'rowid': self.lastrowid,
        #         'source': source_solver
        #     }
        #     db.global_iterations.insert_one(data)

    def record_viewer_data(self, model_viewer_data, key='Driver'):
        """
        Record model viewer data.

        Parameters
        ----------
        model_viewer_data : dict
            Data required to visualize the model.
        key : str, optional
            The unique ID to use for this data in the table.
        """
        pass
        # json_data = json.dumps(model_viewer_data, default=make_serializable)
        # with self.mongodb() as db:
        #     data = {
        #         'case_id': self.case_id,
        #         'id': key,
        #         'model_viewer_data': json_data
        #     }
        #     db.driver_metadata.insert_one(data)

    def record_metadata_system(self, system, run_number=None):
        """
        Record system metadata.

        Parameters
        ----------
        system : System
            The System for which to record metadata.
        run_number : int or None
            Number indicating which run the metadata is associated with.
            None for the first run, 1 for the second, etc.
        """
        pass
        # if self._record_metadata and self.metadata_connection:

        #     scaling_vecs, user_options = self._get_metadata_system(system)

        #     if scaling_vecs is None:
        #         return

        #     path = system.pathname
        #     if not path:
        #         path = 'root'

        #     if run_number is None:
        #         name = path
        #     else:
        #         name = META_KEY_SEP.join([path, str(run_number)])

        #     with self.mongodb() as db:
        #         data = {
        #             'case_id': self.case_id,
        #             'name': name,
        #             'scaling_factors': scaling_vecs,
        #             'component_metadata': user_options
        #         }
        #         db.system_metadata.inser_one(data)

    def record_metadata_solver(self, solver, run_number=None):
        """
        Record solver metadata.

        Parameters
        ----------
        solver : Solver
            The Solver for which to record metadata.
        run_number : int or None
            Number indicating which run the metadata is associated with.
            None for the first run, 1 for the second, etc.
        """
        pass
        # if self._record_metadata and self.metadata_connection:
        #     path = solver._system().pathname
        #     solver_class = type(solver).__name__

        #     if not path:
        #         path = 'root'

        #     id = "{}.{}".format(path, solver_class)

        #     if run_number is not None:
        #         id = META_KEY_SEP.join([id, str(run_number)])

        #     with self.mongodb() as db:
        #         data = {
        #             'case_id': self.case_id,
        #             'id': id,
        #             'solver_class': solver_class
        #         }
        #         db.solver_metadata.inser_one(data)

    def shutdown(self):
        """
        Shut down the recorder.
        """

    def animate_turbineXY(self, duration=10, tail=5, cost='aep', anim_options=None, filename=None):
        """
        Display turbine XY positions between iterations as an animation.

        Parameters
        ----------
        duration : float

        tail : int
            Amount of iterations to show on on the trailing tail.
        cost : str
            name of cost function to be displayed as in MongoDB.
            Defaults to 'aep'.
        filename : str
            Name to save .mp4 animation as.
            Set to "None" to display animation with current Matplotlib backend.
            Defaults to "None".
        Returns
        -------
        ani : Object
            If "filename" exists the method will return a "filename.mp4" file
            Otherwise a Matplotlib animation object is returned.

        """
        x = self['x']
        y = self['y']
        cost = self[cost]
        n_wt = x.shape[1]
        boundary = self['xy_boundary'][-1]
        boundary = np.r_[boundary, boundary[:1]]

        # Setup figure
        fig, ax = plt.subplots()
        # TODO Implement an init() function with axis limits and setting of boundaries
        ax.set_xlim([np.min([np.min(x), np.min(boundary[:, 0])]), np.max([np.max(x), np.max(boundary[:, 0])])])
        ax.set_ylim([np.min([np.min(y), np.min(boundary[:, 1])]), np.max([np.max(y), np.max(boundary[:, 1])])])
        ax.plot(boundary[:, 0], boundary[:, 1])
        # Initialize "ln" list to be updated during animation
        colors = [c['color'] for c in iter(mpl.rcParams['axes.prop_cycle'])] * 100
        ln = [plt.plot([], [], '.-', ms=10, color=c, animated=True)[0] for c in colors[:n_wt]]
        ln += [plt.plot([], [], '--', color=c, animated=True)[0] for c in colors[:n_wt]]
        ln += [plt.plot([], [], 'xk', ms=4, animated=True)[0] for _ in colors[:n_wt]]
        title = ax.text(0.5, .95, "", bbox={'facecolor': 'w', 'alpha': .8, 'pad': 5},
                        transform=ax.transAxes, ha="center")
        ln += [title]

        def update(frame):
            """
            Update data in "ln" for every frame in FuncAnimation().
            """
            title.set_text("%f (%.2f%%)" % (cost[frame],
                                            (cost[frame] - cost[0]) / cost[0] * 100))
            for i in range(n_wt):
                ln[i].set_data(x[max(0, frame - tail):frame + 1, i], y[max(0, frame - tail):frame + 1, i])
                ln[i + n_wt].set_data(np.r_[x[0, i], x[frame, i]], np.r_[y[0, i], y[frame, i]])
                ln[i + 2 * n_wt].set_data(x[frame, i], y[frame, i])
            return ln

        # Run animation
        if anim_options is None:
            ani = animation.FuncAnimation(fig, update, blit=True, frames=x.shape[0], interval=100, repeat=False)
        else:
            ani = animation.FuncAnimation(fig, update, frames=x.shape[0], **anim_options)

        # Exit behaviour decided by presence of a "filename"
        if filename:
            Writer = animation.writers['ffmpeg']
            writer = Writer(fps=len(x) / duration, metadata=dict(artist='TopFarm2', title='Turbine XY - Animation'), bitrate=1800)
            os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)
            ani.save(filename + ".mp4", writer=writer)
        else:
            return ani
